static char rcsid[] = "@(#)$Id: smtp.c,v 2.32 2016/10/18 17:52:59 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.32 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                  (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elmsmtp.h"

#include "../../hdrs/s_elm.h"

#ifdef I_SYSEXITS
#include <sysexits.h>
#endif

DEBUG_VAR(Debug,__FILE__,"smtp");

#ifdef REMOTE_MBX

static const int cancel_delay_msec = 500; /* 500 milliseconds: 1/2 seconds */

#if ANSI_C
#define S_(x) static x;
#define E_(x) extern x;
#else
#define S_(x)
#define E_(x)
#endif

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

enum server_s { Submission_s,
		SMTP_s ,
		submission_s,
		
		NUM_s
};

static const struct string * Server_s P_((enum server_s idx));

#define SMTP_TIMEOUT  240

/* callbacks for connection ----------------------------------------------- */

S_(f_smtp_Server smtp_Server)
static const struct string *smtp_Server  P_((struct mailer_info_1 *I));

S_(ss_action_routine smtp_read_action)
static int smtp_read_action P_((struct streamsched *ss, void * data));
static int smtp_read_action(ss,data) 
     struct streamsched *ss; 
     void * data;
{
    int n;
    struct mailer_info_1 *I;
    int linelen;
    int close_it = 0;

    char * ERROR = NULL;
   
    if (!data)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_read_action",
	      "NULL data argument !!!",0);	
    I = data;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_read_action",
	      "Bad magic number",0);

    if (I->RA.stream != ss) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_read_action",
	      "Bad data argument !!!",0);
    }

    DPRINT(Debug,13,(&Debug,
		     "smtp_read_action: mailer info, I=%p\n",I));

    if (smtp_error == I->state) {
        DPRINT(Debug,13,(&Debug,
                         "smtp_read_action=0 (stream need closing)\n"));
        return 0;
    }

    n = ReadFromStream(ss, &(I->read_buffer),-1);

    if (n < 0) {      
	ERROR = RemoveStreamError(ss);
	DPRINT(Debug,13,(&Debug,
			 "smtp_read_stream: read error %s\n",
			 ERROR ? ERROR : "(none)"));
    }

    DPRINT(Debug,50,(&Debug,
		     "Buffer content="));
#ifdef DEBUG
    if (Debug.active >= 50) {
	int i;

	for (i = 0; i < I->read_buffer.read_len; i++) {
	    DPRINT(Debug,50,(&Debug,
			     " %d",
			     I->read_buffer.read_buffer[i]));
	}
	DPRINT(Debug,50,(&Debug,
			 "\n (message?)="));
	for (i = 0; i < I->read_buffer.read_len; i++) {
	    DPRINT(Debug,50,(&Debug,
			     " %c",
			     isascii(I->read_buffer.read_buffer[i]) &&
			     isprint(I->read_buffer.read_buffer[i]) ? 
			     I->read_buffer.read_buffer[i] :
			     '.'));
	}
    }
#endif
    DPRINT(Debug,50,(&Debug,
		     "\n"));

    while (0 < (linelen = find_crlf(&(I->read_buffer),1))) {
	char code[3];
	int final = 1;
	int i;
	char * message = NULL;
	enum smtp_state old_state = I->state;  /* Handle CLOSING case */

	DPRINT(Debug,13,(&Debug,
                         "smtp_read_action: Parsing response line (len =%d)\n",
			 linelen));

	for (i = 0; i < linelen && i < 3; i++)
	    code[i] = I->read_buffer.read_buffer[i];

	if (i < 3) {
	    int z UNUSED_VAROK = i;

	    for (; i < 3; i++)
		code[i] = '\0';
	
	    DPRINT(Debug,1,(&Debug,
			     "smtp_read_action: Line not include full repsonse code: %.3s (only %d chars)\n",
			    code,z));
	}

	if (linelen > 3)
	    final = I->read_buffer.read_buffer[3] != '-';

	if (linelen > 4)
	    message = & (I->read_buffer.read_buffer[4]);

	DPRINT(Debug,14,(&Debug,
                         "smtp_read_action: Parsing response line, code=%.3s final=%d message=%s\n",
			 code,final,
			 message ? message : "<NULL>"));


	if (0 == memcmp(code,"421",3)) {       /* CLOSING */
	    I->state = smtp_closing;
	    if (message)
		lib_transient(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpServerMsg,
				      "%S server: %s"),
			      smtp_Server(I),
			      message);
	    else
		lib_transient(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpServerClosing,
				      "%S server is closing"),
			      smtp_Server(I));
	} 

	/* Parse response ... */
	if (smtp_wait_reply       == old_state ||
	    smtp_reply_incomplete == I->state) {

	    DPRINT(Debug,13,(&Debug,
			    "smtp_read_action: Waited response\n"));

	    if (smtp_wait_reply       == old_state) {
		if (smtp_wait_reply   == I->state)
		    I->state = smtp_reply_incomplete;
		memcpy(I->response_code,code,3);
		
		if (code[0] < '2' || code[0] > '5') {
		    I->state = smtp_error;
		    lib_error(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpBadResponse,
				      "Bad response from %S server: %s"),
			      Server_s((I->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s),
			      I->read_buffer.read_buffer);
		} 
	    }

	    if (smtp_reply_incomplete == I->state) {
		if (0 != memcmp(I->response_code,code,3)) {
		    I->state = smtp_error;
		    lib_error(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpBadResponse,
				      "Bad response from %S server: %s"),
			      Server_s((I->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s),
			      I->read_buffer.read_buffer);
		}

		if (final) {
		    if ('3' == code[0])
			I->state = smtp_need_data;
		    else
			I->state = smtp_reply_read;
		    
		    DPRINT(Debug,13,(&Debug,
				     "smtp_read_action : %sresponse=%.3s %s\n",
				     smtp_reply_read == I->state ?
				     "command ready, " : "",
				     I->response_code,
				     message ? message : "(no message text)"));
		}
	    }
	    
	    if (message) {
		I->response_text = safe_array_realloc(I->response_text,
						      (I->response_text_len+1),
						      sizeof (*(I->response_text)));
		I->response_text[I->response_text_len++] = 
		    safe_strdup(message);
	    }
	    
	    cut_line(&(I->read_buffer),linelen);
	    message = NULL;          /* No longer valid */
	} else {
	    DPRINT(Debug,16,(&Debug,
			     "smtp_read_action: Quiting parsing loop (%d bytes left, unaccepted lines was %d bytes)\n",
			     I->read_buffer.read_len,
			     linelen));
	    break;
	}
    }

    if (I->read_buffer.read_len) {
	switch(I->state) {
	    int i;
	case smtp_reply_read:
	case smtp_need_data:
	case smtp_data:
	case smtp_idle:
	case smtp_command:
	    DPRINT(Debug,13,(&Debug,
			     "smtp_read_action: %d bytes data left! (bad state):\n",
			     I->read_buffer.read_len));
	    DPRINT(Debug,13,(&Debug,
			     "smtp_read_action: extra data="));
	    for (i =0; i < I->read_buffer.read_len; i++) {
		DPRINT(Debug,13,(&Debug,
				 " %d",
				 I->read_buffer.read_buffer[i]));
	    }
	    DPRINT(Debug,13,(&Debug,
			     "\nsmtp_read_action (message?)="));
	    for (i =0; i < I->read_buffer.read_len; i++) {
		DPRINT(Debug,13,(&Debug,
				 "%c",
				 isascii(I->read_buffer.read_buffer[i]) &&
				 isprint(I->read_buffer.read_buffer[i]) ?
				 I->read_buffer.read_buffer[i] :
				 '.'));
	    }
	    DPRINT(Debug,13,(&Debug,
			     "\n"));
	    
	    lib_error(CATGETS(elm_msg_cat, SmtpSet,
			      SmtpUnexpectedResponse,
			      "Unexpected response from %S server"),
		      Server_s((I->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s));
	    close_it = 1;
	    break;

	default:
	    break;
	}
    }

    if (n == 0) {
	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpConnectionClosed,
			  "Connection closed to %S server"),
		  Server_s((I->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s));
	close_it = 1;
    }
    if (n < 0 && ERROR) {
	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpErrorReading,
			  "Error reading from %S server: %s"),
		  Server_s((I->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s),
		  ERROR);
	close_it = 1;
    }
    if (ERROR) {
	free(ERROR);
	ERROR = NULL;
    }
    if (close_it) {
	I->state = smtp_error;
	DPRINT(Debug,13,(&Debug,
			 "smtp_read_action=0 (stream need closing)\n"));
	return 0;
    }
    DPRINT(Debug,13,(&Debug,
		     "smtp_read_action=1\n"));

    return 1;
}

S_(ss_action_routine smtp_write_action)
static int smtp_write_action P_((struct streamsched *ss, void * data));
static int smtp_write_action(ss,data) 
     struct streamsched *ss; 
     void * data;
{
    char * ERROR = NULL;
    struct mailer_info_1 *I;
    int n;

    if (!data)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_write_action",
	      "NULL data argument !!!",0);	
    I = data;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_write_action",
	      "Bad magic number",0);

    if (I->RA.stream != ss) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_write_action",
	      "Bad data argument !!!",0);
    }

    DPRINT(Debug,13,(&Debug,
		     "smtp_write_action: mailer info, I=%p\n",I));

    if (!I->write_buffer.write_len) {
	DPRINT(Debug,13,(&Debug,
			 "smtp_write_action: NO DATA TO WRITE!\n"));
    }

    if (smtp_error   == I->state ||
	smtp_closing == I->state) {
	DPRINT(Debug,13,(&Debug,
			 "smtp_write_action=0 (stream need closing)\n"));
	return 0;
    }

    if (smtp_command != I->state &&
	smtp_data    != I->state) {

	DPRINT(Debug,1,(&Debug,
			"smtp_write_action: state=%d\n",
			I->state));

	panic("MBX PANIC",__FILE__,__LINE__,"smtp_write_action",
	      "Bad state !!!",0);
    }

    n = WriteToStream(ss, & (I->write_buffer));

    if (n > 0) {
	int i;
	
	DPRINT(Debug,70,(&Debug,
			 "Write content="));
	for (i = 0; i < n; i++) {
	    DPRINT(Debug,70,(&Debug," %d",
			     I->write_buffer.write_buffer[i]));
	}
	DPRINT(Debug,70,(&Debug,"\n    (text?)="));
	for (i = 0; i < n; i++) {
	    DPRINT(Debug,70,(&Debug," %c",
			     isascii(I->write_buffer.write_buffer[i]) 
			     &&
			     isprint(I->write_buffer.write_buffer[i]) 
			     ? 
			     I->write_buffer.write_buffer[i] :
			     '.'));
	}
	DPRINT(Debug,70,(&Debug,"\n"));
	
	cut_Write_Buffer(& (I->write_buffer),n);

    } else if (n < 0) {
	ERROR = RemoveStreamError(ss);
	
	DPRINT(Debug,13,(&Debug,
			 "smtp_write_action: write error %s\n",
			 ERROR ? ERROR : "(none)"));
    }

    if (!I->write_buffer.write_len) {
	if (smtp_command == I->state)
	    I->state = smtp_wait_reply;
	else
	    I->state = smtp_need_data;

	if (ERROR) {
	    free(ERROR);
	    ERROR = NULL;
	}
	       
	DPRINT(Debug,13,(&Debug,
			 "smtp_write_action=0 -- data written\n"));
	return 0;
    }
    
    if (ERROR) {
	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpErrorWriting,
			  "Error writing to %S server: %s"),
		  Server_s((I->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s),
		  ERROR);
	I->state = smtp_error;
	DPRINT(Debug,13,(&Debug,
			 "smtp_write_action=0 (stream need closing)\n"));

	free(ERROR);
	ERROR = NULL;
        return 0;	
    }

    DPRINT(Debug,13,(&Debug,
		     "smtp_write_action=1\n"));
    return 1;

}

S_(f_smtp_start_command smtp_start_command)
static int smtp_start_command P_((struct mailer_info_1 *I, char *command));

S_(f_smtp_command_ready smtp_command_ready)
static void smtp_command_ready P_((struct mailer_info_1 *I, int data_end));

S_(ss_action_routine smtp_timeout_action)
static int smtp_timeout_action P_((struct streamsched *ss, void * data));
static int smtp_timeout_action(ss,data) 
     struct streamsched *ss; 
     void * data;
{
    struct mailer_info_1 *I;
    int status;

    if (!data)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_timeout_action",
	      "NULL data argument !!!",0);	
    I = data;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_timeout_action",
	      "Bad magic number",0);

    if (I->RA.stream != ss) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_timeout_action",
	      "Bad data argument !!!",0);
    }

    DPRINT(Debug,13,(&Debug,
		     "smtp_timeout_action: mailer info, I=%p\n",I));

    if (smtp_error   == I->state ||
	smtp_closing == I->state) {
	DPRINT(Debug,13,(&Debug,
			 "smtp_timeout_action=0 (stream need closing)\n"));
	return 0;
    }

    if (smtp_idle != I->state) {
	DPRINT(Debug,13,(&Debug,
			 "smtp_timeout_action=1 (stream not idle)\n"));
	return 1;
    }

    status = smtp_start_command(I,"NOOP");

    if (status) 
	smtp_command_ready(I,0);
    /* DO not wait completion of command */

    DPRINT(Debug,13,(&Debug,
                     "smtp_timeout_action=%d\n",
                     status));

    return status;
}

static void free_response_text P_((struct mailer_info_1 *I));
static void free_response_text(I)
     struct mailer_info_1 *I;
{
    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"free_response_text",
	      "Bad magic number",0);

    if (I->response_text) {
	int i;
	for (i = 0; i < I->response_text_len; i++) {
	    if (I->response_text[i]) {
		free(I->response_text[i]);
		I->response_text[i] = NULL;
	    }
	}
	free(I->response_text);
    }            
    I->response_text     = NULL;
    I->response_text_len = 0;
}

static int smtp_start_command_tail P_((struct mailer_info_1 *I,
				       char *command));
static int smtp_start_command_tail(I,command)
     struct mailer_info_1 *I; 
     char *command;
{
    int status = 0;

    if (!I->RA.stream) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_start_command_tail: stream not open\n"));
	status = 0;
	goto clean;
    }
    
    if (I->incomplete_command)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_start_command_tail",
	      "incomplete_command not NUL !",0);
    I->incomplete_command_len = 0;

    I->state = smtp_command;

    DPRINT(Debug,15,(&Debug,
		     "smtp_start_command_tail: Command %.4s...\n",command));

    I->incomplete_command     = safe_strdup(command);
    I->incomplete_command_len = strlen(command);
    status = 1;

 clean:
    DPRINT(Debug,15,(&Debug,
		     "smtp_start_command_tail=%d\n",status));
    return status;
}

static int smtp_start_command_afterwait P_((struct mailer_info_1 *I));
static int smtp_start_command_afterwait(I)
     struct mailer_info_1 *I;
{
    int status = 0;

    /* DO NOT accpet smtp_need_data here -- it means that
       data (not command) is expected
    */
    if (smtp_error == I->state ||
	smtp_closing == I->state) {
	
	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_start_command_afterwait: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}	
	status = 0;
	goto clean;
    }

    if (0 != I->write_buffer.write_len) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_start_command_afterwait: write buffer not empty (%d bytes left)\n",
			 I->write_buffer.write_len));
	status = 0;
	goto clean;
    }       	
    
    free_response_text(I);
    
    status = 1;

clean:
    DPRINT(Debug,15,(&Debug,
		     "smtp_start_command_afterwait=%d\n",status));
    return status;
}

static int smtp_start_command_init P_((struct mailer_info_1 *I));
static int smtp_start_command_init(I)
     struct mailer_info_1 *I;
{
    int status = 0;

    if (smtp_error == I->state   ||
	smtp_closing == I->state ||
	!I->RA.stream) {
	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_start_command_init: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}	
	status = 0;
	goto clean;
    }
    
    status = 1;

 clean:
    DPRINT(Debug,15,(&Debug,
		     "smtp_start_command_init=%d\n",status));
    return status;
}   

static int smtp_start_command(I,command)
     struct mailer_info_1 *I; 
     char *command;
{
    int status = 0;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_start_command",
	      "Bad magic number",0);
    
    DPRINT(Debug,12,(&Debug,
		     "smtp_start_command: mailer info, I=%p\n",
		     I));
    
    if (!smtp_start_command_init(I)) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_start_command: failed on init\n"));
	status = 0;
	goto clean;
    }

    if (I->state != smtp_idle) {
	DPRINT(Debug,7,(&Debug,
			"smtp_start_command: wrong state -- perhaps idle timer -- waiting...\n"));
	
	while (smtp_reply_read != I->state &&
	       smtp_need_data    != I->state &&
	       smtp_error        != I->state &&
	       smtp_closing      != I->state &&
	       NULL         != I->RA.stream) {
	    WaitStreamFor(I->RA.stream,SS_read_act);
	    /* Should user to have some way to interrupt progress ? 
	       cancelable version is smtp_start_command_c()
	    */
	}
	DPRINT(Debug,7,(&Debug,
			 "smtp_start_command: ... wait ended\n"));

	if (!smtp_start_command_afterwait(I)) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_start_command: failed after wait\n"));
	    status = 0;
	    goto clean;
	}
    }

    if (!smtp_start_command_tail(I,command)) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_start_command: failure\n"));
	status = 0;
	goto clean;
    }

    status = 1;

 clean:
    DPRINT(Debug,12,(&Debug,
		     "smtp_start_command=%d\n",
		     status));

    return status;
}

static void smtp_cancel_reset_stream P_((struct mailer_info_1 *I));
static void smtp_cancel_reset_stream(I)
     struct mailer_info_1 *I;
{
    if ((smtp_error != I->state &&
	 smtp_closing != I->state) 
	||
	I->RA.stream) {

	DPRINT(Debug,12,(&Debug,
			 "smtp_cancel_reset_stream: Stream no longer usable because of cancel\n"));

	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_cancel_reset_stream: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}
	
	if (smtp_error != I->state &&
	    smtp_closing != I->state) {
	    I->state = smtp_error;

	    DPRINT(Debug,12,(&Debug,
			     "smtp_cancel_reset_stream: Stream on error state\n"));
	}	
    }
}



static int smtp_start_command_c P_((struct mailer_info_1 *I, 
				    char *command,
				    struct cancel_data *cd));
static int smtp_start_command_c(I,command,cd)
     struct mailer_info_1 *I;
     char *command;
     struct cancel_data *cd;
{
    int status = 0;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_start_command",
	      "Bad magic number",0);
    
    DPRINT(Debug,12,(&Debug,
		     "smtp_start_command_c: mailer info, I=%p\n",
		     I));

    if (!smtp_start_command_init(I)) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_start_command_c: failed on init\n"));
	status = 0;
	goto clean;
    }

    if (I->state != smtp_idle) {
	DPRINT(Debug,7,(&Debug,
			"smtp_start_command_c: wrong state -- perhaps idle timer -- waiting...\n"));
	while (smtp_reply_read != I->state &&
	       smtp_need_data    != I->state &&
	       smtp_error        != I->state &&
	       smtp_closing      != I->state &&
	       NULL         != I->RA.stream) {

	    if (!WaitStreamFor_c(I->RA.stream,SS_read_act,cd)) {
		
		DPRINT(Debug,7,(&Debug,"smtp_start_command_c: wait not succeed?\n"));
		
		if (cd && is_canceled(cd)) {
		    DPRINT(Debug,7,(&Debug,"smtp_start_command_c: wait canceled!\n"));
		    
		    status = 0;
		    goto clean;
		}
	    }
	}

	DPRINT(Debug,7,(&Debug,
			 "smtp_start_command_c: ... wait ended\n"));

	if (!smtp_start_command_afterwait(I)) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_start_command: failed after wait\n"));
	    status = 0;
	    goto clean;
	}
   }

    if (!smtp_start_command_tail(I,command)) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_start_command_c: failure\n"));
	status = 0;
	goto clean;
    }

    status = 1;

clean:
    DPRINT(Debug,12,(&Debug,
		     "smtp_start_command_c=%d\n",
		     status));

    return status;

}

static void smtp_push_data_add_dataline P_((struct mailer_info_1 *I,
					   const char *data,
					   int len));
static void smtp_push_data_add_dataline(I,data,len)
     struct mailer_info_1 *I;
     const char *data;
     int len;
{

    char * d,* d1;
    int len1;

    /* Dot-Stuff data line */
    if ('.' == data[0]) {
	len1 = len + 1;

	d = safe_malloc(len1);
	d[0] = '.';
	d1 = d + 1;
    } else {
	d = safe_malloc(len);
	d1 = d;
	len1 = len;
    }

    memcpy(d1,data,len);
    add_to_Write_Buffer(& (I->write_buffer),&d,len1);   /* May free d */

    ConfigStream(I->RA.stream,
		 smtp_read_action,smtp_write_action,smtp_timeout_action,
		 SMTP_TIMEOUT,I);    
}

static void smtp_push_data_add_command P_((struct mailer_info_1 *I,
					   const char *data,
					   int len));
static void smtp_push_data_add_command(I,data,len)
     struct mailer_info_1 *I;
     const char *data;
     int len;
{
    if (I->incomplete_command)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_push_data_add_command",
	      "incomplete_command not NUL !",0);

    I->incomplete_command = safe_malloc(len);
    memcpy(I->incomplete_command,data,len);
    I->incomplete_command_len = len;  
}

static int smtp_push_data_init P_((struct mailer_info_1 *I));
static int smtp_push_data_init(I)
     struct mailer_info_1 *I;
{
    int status = 0;

    if (smtp_error == I->state   ||
	smtp_closing == I->state ||
	!I->RA.stream) {

	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_push_data_init: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}	
	status = 0;
	goto clean;
    }

    if (I->incomplete_command)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_push_data_init",
	      "incomplete_command not NUL !",0);
    I->incomplete_command_len = 0;

        if (smtp_need_data != I->state  &&
	smtp_data != I->state) {
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_push_data_init",
	      "Wrong state !",0);
    }
    I->state = smtp_data;

    status = 1;

 clean:
    DPRINT(Debug,15,(&Debug,
		     "smtp_push_data_init=%d\n",status));
    return status;
}

S_(f_smtp_push_data smtp_push_data)
static int smtp_push_data P_((struct mailer_info_1 *I,
			      const char *data,
			      int len,
			      int is_data_line));

static int smtp_push_data(I,data,len,is_data_line)
     struct mailer_info_1 *I; 
     const char *data;
     int len;
     int is_data_line;
{
    int status = 0;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_push_data",
	      "Bad magic number",0);
    
    DPRINT(Debug,12,(&Debug,
		     "smtp_push_data: mailer info, I=%p\n",
		     I));

    if (!smtp_push_data_init(I)) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_push_data: failed on init\n"));
	status = 0;
	goto clean;
    }
    
    if (is_data_line) {

	smtp_push_data_add_dataline(I,data,len);

	/* We push large amout of data so start immediately without
	   buffering all of data

	   Cancelable version is smtp_push_data_c()
	*/
	WaitStreamFor(I->RA.stream,SS_write_act);
    } else 
	smtp_push_data_add_command(I,data,len);

    status = 1;

 clean:
    DPRINT(Debug,12,(&Debug,
		     "smtp_push_data=%d\n",status));
    return status;
}

static int smtp_push_data_c P_((struct mailer_info_1 *I,
				const char *data,
				int len,
				int is_data_line,
				struct cancel_data *cd))
    UNUSED_FUNCOK;
static int smtp_push_data_c(I,data,len,is_data_line,cd)
     struct mailer_info_1 *I; 
     const char *data;
     int len;
     int is_data_line;
     struct cancel_data *cd;
{   
    int status = 0;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_push_data_c",
	      "Bad magic number",0);
    
    DPRINT(Debug,12,(&Debug,
		     "smtp_push_data_c: mailer info, I=%p\n",
		     I));

    if (!smtp_push_data_init(I)) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_push_data_c: failed on init\n"));
	status = 0;
	goto clean;
    }

    if (is_data_line) {

	smtp_push_data_add_dataline(I,data,len);

	/* We push large amout of data so start immediately without
	   buffering all of data
	*/

	if (!WaitStreamFor_c(I->RA.stream,SS_write_act,cd)) {
	    DPRINT(Debug,7,(&Debug,"smtp_push_data_c: wait not succeed?\n"));

	    if (cd && is_canceled(cd)) {
		DPRINT(Debug,7,(&Debug,"smtp_push_data_c: wait canceled!\n"));
		
		smtp_cancel_reset_stream(I);
		status = 0;
		goto clean;
	    }
	}

    } else 
	smtp_push_data_add_command(I,data,len);

    status = 1;

 clean:
    DPRINT(Debug,12,(&Debug,
		     "smtp_push_data_c=%d\n",status));
    return status;
}


S_(f_smtp_command_arg smtp_command_arg)
static void smtp_command_arg P_((struct mailer_info_1 *I, const char *str));
static void smtp_command_arg(I, str)
     struct mailer_info_1 *I;
     const char *str;
{
    int L = strlen(str);

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_arg",
	      "Bad magic number",0);

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_arg: mailer info, I=%p\n",
		     I));

    if (I->state != smtp_command &&
	I->state != smtp_data) {

	DPRINT(Debug,1,(&Debug,
			"smtp_command_arg: state=%d\n",
			I->state));

	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_arg",
	      "Bad state",0);
    }

    I->incomplete_command = safe_realloc(I->incomplete_command,
					 I->incomplete_command_len+1+L);
    I->incomplete_command[I->incomplete_command_len++] = ' ';

    memcpy(I->incomplete_command+I->incomplete_command_len,str,L);
    I->incomplete_command_len += L;
}

static void smtp_command_ready_init P_((struct mailer_info_1 *I, int data_end));
static void smtp_command_ready_init(I,data_end) 
     struct mailer_info_1 *I;
     int data_end;
{
    char *s;

    if (I->state != smtp_command &&
	I->state != smtp_data) {
	
	DPRINT(Debug,1,(&Debug,
			"smtp_command_ready_init: state=%d\n",
			I->state));

	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_ready_init",
	      "Bad state",0);
    }
    
    if (data_end && I->state != smtp_data) {
	
	DPRINT(Debug,1,(&Debug,
			"smtp_command_ready_init: state=%d\n",
			I->state));

	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_ready_init",
	      "Bad state -- not smtp data",0);
    }
    
    /* May free I->p.p1->incomplete_command */
    add_to_Write_Buffer(& (I->write_buffer),
			&(I->incomplete_command),
			I->incomplete_command_len);
    
    s = safe_strdup("\r\n");
    /* May free s */
    add_to_Write_Buffer(& (I->write_buffer),&s,2);

    ConfigStream(I->RA.stream,
		 smtp_read_action,smtp_write_action,smtp_timeout_action,
		 SMTP_TIMEOUT,I);

    
}


static void smtp_command_ready(I, data_end)
     struct mailer_info_1 *I;
     int data_end;
{

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_ready",
	      "Bad magic number",0);

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ready: mailer info, I=%p\n",
		     I));
    
    smtp_command_ready_init(I, data_end) ;

    if (data_end) {
	I->state = smtp_command;
	/* We push large amout of data so start immediately without
	   buffering all of data

	   Cancelable version is smtp_command_ready_c()
	*/
	WaitStreamFor(I->RA.stream,SS_write_act);
    }

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ready: done%s\n",
		     smtp_command == I->state ? " (state: smtp_command)" : ""
		     ));
}

static int smtp_command_ready_c P_((struct mailer_info_1 *I,
				     int data_end,
				    struct cancel_data *cd));
static int smtp_command_ready_c(I,data_end,cd)
     struct mailer_info_1 *I;
     int data_end;
     struct cancel_data *cd;
{
    int status = 0;
    
    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_ready_c",
	      "Bad magic number",0);
    
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ready_c: mailer info, I=%p\n",
		     I));
    
    smtp_command_ready_init(I, data_end) ;
    
    if (data_end) {
	I->state = smtp_command;
	/* We push large amout of data so start immediately without
	   buffering all of data
	*/

	if (!WaitStreamFor_c(I->RA.stream,SS_write_act,cd)) {
	    DPRINT(Debug,7,(&Debug,"smtp_command_ready_c: wait not succeed?\n"));

	    if (cd && is_canceled(cd)) {
		DPRINT(Debug,7,(&Debug,"smtp_command_ready_c: wait canceled!\n"));
		
		smtp_cancel_reset_stream(I);
		status = 0;
		goto clean;
	    }
	}
    }

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ready_c: done%s\n",
		     smtp_command == I->state ? " (state: smtp_command)" : ""
		     )); 

    status = 1;

 clean:
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ready_c=%d\n",status));
    return status;

}
				   
static int smtp_command_ok_tail P_((struct mailer_info_1 *I, 
				    char response_code[4]));
static int smtp_command_ok_tail(I,response_code)
     struct mailer_info_1 *I;
     char response_code[4];
{
    int status = 0;

    if (smtp_reply_incomplete == I->state ||
	smtp_reply_read     == I->state ||
	smtp_need_data        == I->state) {
	memcpy(response_code,I->response_code,3);
	
	if ('2' == response_code[0] ||
	    '3' == response_code[0])
	    status = 1;

	response_code[3] = '\0';
    }

    DPRINT(Debug,15,(&Debug,
		     "smtp_command_ok_tail=%d\n",status));
    return status;
}

static int smtp_command_ok_afterwait_r P_((struct mailer_info_1 *I,
					 char response_code[4]));
static int smtp_command_ok_afterwait_r(I,response_code)
     struct mailer_info_1 *I;
     char response_code[4];
{
    int status = 0;

    if (smtp_error == I->state ||
	smtp_closing == I->state) {
	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_ok_afterwait_r: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}	
	status = 0;
	goto clean;
    }
    
    if (!I->RA.stream) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_command_ok_afterwait_r: stream not open\n"));
	status = 0;
	goto clean;
    }

    status = 1;
    
 clean:
    DPRINT(Debug,15,(&Debug,
		     "smtp_command_ok_afterwait_r=%d\n",status));
    return status;
}


static int smtp_command_ok_afterwait_w P_((struct mailer_info_1 *I,
					 char response_code[4]));
static int smtp_command_ok_afterwait_w(I,response_code)
     struct mailer_info_1 *I;
     char response_code[4];
{
    int status = 0;

    if (smtp_error == I->state ||
	smtp_closing == I->state) {

	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_ok_afterwait_w: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}	
	status = 0;
	goto clean;
    }

    status = 1;
    
 clean:
    DPRINT(Debug,15,(&Debug,
		     "smtp_command_ok_afterwait_w=%d\n",status));
    return status;
}


static int smtp_command_ok_init P_((struct mailer_info_1 *I, 
				    char response_code[4]));
static int smtp_command_ok_init(I,response_code)
     struct mailer_info_1 *I;
     char response_code[4];
{
    int status = 0;

    response_code[0] = '\0';
    
    if (smtp_error == I->state   ||
	smtp_closing == I->state ||
	!I->RA.stream) {

	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_ok_init: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}	
	status = 0;
	goto clean;
    }

    if (smtp_command          != I->state &&
	smtp_reply_read     != I->state &&
	smtp_need_data        != I->state &&
	smtp_reply_incomplete != I->state &&
	smtp_wait_reply       != I->state) {
	
	DPRINT(Debug,1,(&Debug,
			"smtp_command_ok_init: state=%d\n",
			I->state));
	
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_ok_init",
	      "Bad state",0);
    }

    if (I->incomplete_command)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_ok_init",
	      "incomplete_command not NULL -- smtp_command_ready() not called",0);

    status = 1;

 clean:
    DPRINT(Debug,15,(&Debug,
		     "smtp_command_ok_init=%d\n",status));
    return status;
}


S_(f_smtp_command_ok smtp_command_ok)
/* May return command OK before all response lines are returned */
static int smtp_command_ok P_((struct mailer_info_1 *I, 
			       char response_code[4]));
static int smtp_command_ok(I,response_code)
     struct mailer_info_1 *I;
     char response_code[4];
{
    int status = 0;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_ok",
	      "Bad magic number",0);

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ok: mailer info, I=%p\n",
		     I));

    if (!smtp_command_ok_init(I,response_code)) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_start_command_init: failed on init\n"));
	status = 0;
	goto clean;
    }
   
    if (smtp_command     == I->state) {
	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok: writing data to server...\n"));
	
	while ( smtp_command  == I->state &&
		NULL         != I->RA.stream) {
	    WaitStreamFor(I->RA.stream,SS_write_act);
	    /* Should user to have some way to interrupt progress ? 
	       Cancelable version is smtp_command_ok_c()
	     */
	}
	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok: ... write ended\n"));

	if (! smtp_command_ok_afterwait_w(I,response_code)) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_ok: failure after wait (write)\n"));
	    status = 0;
	    goto clean;
	}
    }

    /* If nothing read, wait data for now */
    if (smtp_wait_reply  == I->state) {
      
	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok: waiting reply from server...\n"));

	while ( smtp_wait_reply  == I->state &&
		NULL         != I->RA.stream) {
	    WaitStreamFor(I->RA.stream,SS_read_act);
	    /* Should user to have some way to interrupt progress ? 
	       Cancelable version is smtp_command_ok_c()
	     */
	}
	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok: ... wait ended\n"));

	if (! smtp_command_ok_afterwait_r(I,response_code)) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_ok: failure after wait (reply)\n"));
	    status = 0;
	    goto clean;
	}
    }

    if (! smtp_command_ok_tail(I,response_code)) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_start_command_ok: failure\n"));
	status = 0;
	goto clean;
    }

    status = 1;
    
 clean:
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ok=%d\n",status));
    return status;
}

/* May return command OK before all response lines are returned */
static int smtp_command_ok_c P_((struct mailer_info_1 *I, 
				 char response_code[4],
				 struct cancel_data *cd));
static int smtp_command_ok_c(I,response_code,cd)
     struct mailer_info_1 *I;
     char response_code[4];
     struct cancel_data *cd;
{
    int status = 0;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_ok_c",
	      "Bad magic number",0);
    
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ok_c: mailer info, I=%p\n",
		     I));
    
    if (!smtp_command_ok_init(I,response_code)) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_start_command_init_c: failed on init\n"));
	status = 0;
	goto clean;
    }

    if (smtp_command     == I->state) {
	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok_c: writing data to server...\n"));

	while ( smtp_command  == I->state &&
		NULL         != I->RA.stream) {
	
	    if (!WaitStreamFor_c(I->RA.stream,SS_write_act,cd)) {
		DPRINT(Debug,7,(&Debug,"smtp_command_ok_c: wait not succeed?\n"));
		
		if (cd && is_canceled(cd)) {
		    DPRINT(Debug,7,(&Debug,"smtp_command_ok_c: wait canceled!\n"));
		    
		    smtp_cancel_reset_stream(I);
		    status = 0;
		    goto clean;
		}
	    }
	}

	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok_c: ... write ended\n"));

	if (! smtp_command_ok_afterwait_w(I,response_code)) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_ok_c: failure after wait (write)\n"));
	    status = 0;
	    goto clean;
	}	
    }
    
    /* If nothing read, wait data for now */
    if (smtp_wait_reply  == I->state) {
      
	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok_c: waiting reply from server...\n"));

	while ( smtp_wait_reply  == I->state &&
		NULL         != I->RA.stream) {

	    if (!WaitStreamFor_c(I->RA.stream,SS_read_act,cd)) {

		DPRINT(Debug,7,(&Debug,"smtp_command_ok_c: wait not succeed?\n"));

		if (cd && is_canceled(cd)) {
		    DPRINT(Debug,7,(&Debug,"smtp_command_ok_c: wait canceled!\n"));
		    
		    smtp_cancel_reset_stream(I);
		    status = 0;
		    goto clean;
		}	       
	    }
	}

	DPRINT(Debug,7,(&Debug,
			"smtp_command_ok_c: ... wait ended\n"));

	if (! smtp_command_ok_afterwait_r(I,response_code)) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_start_command_ok_c: failure after wait\n"));
	    status = 0;
	    goto clean;
	}
    }

    if (! smtp_command_ok_tail(I,response_code)) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_start_command_ok_c: failure\n"));
	status = 0;
	goto clean;
    }
    
    status = 1;
      
 clean:
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_ok_c=%d\n",status));
    return status;
}

static const char * smtp_command_response_tail P_((struct mailer_info_1 *I,
					     int *idx));
static const char * smtp_command_response_tail(I,idx)
     struct mailer_info_1 *I;
     int *idx;
{
    char * status = NULL;

    if (*idx < I-> response_text_len) {
	status = I-> response_text[(*idx)++];
    }


    DPRINT(Debug,15,(&Debug,
		     "smtp_command_ok_init=%s\n",
		     status ? status : "<NULL>"));

    return status;
}

static int smtp_command_response_afterwait 
   P_((struct mailer_info_1 *I));
static int smtp_command_response_afterwait(I)
     struct mailer_info_1 *I;
{
    int status = 0;

    if (smtp_reply_incomplete != I->state &&
	smtp_reply_read     != I->state &&
	smtp_need_data        != I->state) {

	DPRINT(Debug,12,(&Debug,
			 "smtp_command_response_afterwait: Bad state, state=%d\n",
			 I->state));

	status = 0;
	goto clean;
    }
    status = 1;

 clean:
    DPRINT(Debug,15,(&Debug,
		     "smtp_command_response_afterwait=%d\n",status));
    return status;
}

static int smtp_command_response_init P_((struct mailer_info_1 *I,
					  int *idx));
static int smtp_command_response_init(I,idx)
     struct mailer_info_1 *I;
     int *idx;
{
    int status = 0;

    if (*idx < 0) {
	panic("MAILER PANIC",__FILE__,__LINE__,
	      "smtp_command_response_init",
	      "Negative *idx",0);
    }

    if (smtp_reply_incomplete != I->state &&
	smtp_reply_read     != I->state &&
	smtp_need_data        != I->state) {
	
	DPRINT(Debug,12,(&Debug,
			"smtp_command_response_init: bad state, state=%d\n",
			I->state));

	status = 0;
	goto clean;
    }
    
    status = 1;

 clean:
    DPRINT(Debug,15,(&Debug,
		     "smtp_command_response_init=%d\n",status));
    return status;
}   



S_(f_smtp_command_response smtp_command_response)
/* Returns response lines one at time and waits that all is returned */
static const char * smtp_command_response P_((struct mailer_info_1 *I,
					int *idx));

static const char * smtp_command_response(I,idx)
     struct mailer_info_1 *I;
     int *idx;
{
    const char * status = NULL;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_response",
	      "Bad magic number",0);

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_response: mailer info, I=%p, *idx=%d\n",
		     I,*idx));

    if (! smtp_command_response_init(I,idx)) {
	DPRINT(Debug,12,(&Debug,
			     "smtp_command_response: failure on init\n"));
	    status = NULL;
	    goto clean;
    }
        

    if (*idx >= I-> response_text_len &&
	smtp_reply_incomplete == I->state) {
	DPRINT(Debug,7,(&Debug,
			"smtp_command_response: waiting more data from server...\n"));
	while (*idx >= I-> response_text_len &&
	       smtp_reply_incomplete == I->state &&
	       NULL         != I->RA.stream) {
            WaitStreamFor(I->RA.stream,SS_read_act);
            /* Should user to have some way to interrupt progress ? 
	       Cancelable version is smtp_command_response_c()
	     */
        }
	DPRINT(Debug,7,(&Debug,
			"smtp_command_response: ... wait ended\n"));

	if (!smtp_command_response_afterwait(I)) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_response: failure after wait\n"));
	    status = NULL;
	    goto clean;
	}
    }

    status = smtp_command_response_tail(I,idx);

 clean:

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_response=%s\n",
		     status ? status : "<NULL>"));

    return status;
}

/* Returns response lines one at time and waits that all is returned */
static const char * smtp_command_response_c P_((struct mailer_info_1 *I,
						int *idx,
						struct cancel_data *cd));
static const char * smtp_command_response_c(I,idx,cd)
     struct mailer_info_1 *I;
     int *idx;
     struct cancel_data *cd;
{
    const char * status = NULL;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_response_c",
	      "Bad magic number",0);

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_response_c: mailer info, I=%p, *idx=%d\n",
		     I,*idx));

    if (! smtp_command_response_init(I,idx)) {
	DPRINT(Debug,12,(&Debug,
			 "smtp_command_response_c: failure on init\n"));
	status = NULL;
	goto clean;
    }

     if (*idx >= I-> response_text_len &&
	smtp_reply_incomplete == I->state) {
	DPRINT(Debug,7,(&Debug,
			"smtp_command_response_c: waiting more data from server...\n"));
	
	while (*idx >= I-> response_text_len &&
	       smtp_reply_incomplete == I->state &&
	       NULL         != I->RA.stream) {

	    if (!WaitStreamFor_c(I->RA.stream,SS_read_act,cd)) {

		DPRINT(Debug,7,(&Debug,"smtp_command_response_c: wait not succeed?\n"));
		if (cd && is_canceled(cd)) {
		    DPRINT(Debug,7,(&Debug,"smtp_command_response_c: wait canceled!\n"));
		    
		    smtp_cancel_reset_stream(I);
		    status = NULL;
		    goto clean;
		}
	    }

	}

	DPRINT(Debug,7,(&Debug,
			"smtp_command_response_c: ... wait ended\n"));

	if (!smtp_command_response_afterwait(I)) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_response_c: failure after wait\n"));
	    status = NULL;
	    goto clean;
	}
    }

    status = smtp_command_response_tail(I,idx);

 clean:
    
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_response_c=%s\n",
		     status ? status : "<NULL>"));

    return status;
}

static void smtp_command_reset_tail P_((struct mailer_info_1 *I));
static void smtp_command_reset_tail(I)
     struct mailer_info_1 *I;
{
    if (smtp_error == I->state   ||
	smtp_closing == I->state) {
	if (I->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_command_reset_tail: Closing stream\n"));
	    
	    FreeStreamStack( &(I->RA.stream));
	}	
    }
    
    free_response_text(I);

    if (smtp_error     != I->state  &&
	smtp_closing   != I->state &&
	smtp_need_data != I->state)
	I->state = smtp_idle;

}

S_(f_smtp_command_reset smtp_command_reset)
/* Waits that command is completed and resets data then */
static void smtp_command_reset P_((struct mailer_info_1 *I));
static void smtp_command_reset(I)
     struct mailer_info_1 *I;
{

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_reset",
	      "Bad magic number",0);

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_reset: mailer info, I=%p\n",I));
    
    if (smtp_command     == I->state) {
	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset: writing data to server...\n"));
	
	while ( smtp_command  == I->state &&
		NULL         != I->RA.stream) {
	    WaitStreamFor(I->RA.stream,SS_write_act);
	    /* Should user to have some way to interrupt progress ? 
	       Cancelable version is smtp_command_reset_c()
	     */
	}
	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset: ... write ended\n"));
    }

    if (smtp_wait_reply  == I->state ||
	smtp_reply_incomplete == I->state) {

	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset: waiting reply from server...\n"));

	while ( (smtp_wait_reply       == I->state || 
		 smtp_reply_incomplete == I->state) &&
		NULL         != I->RA.stream) {
	    WaitStreamFor(I->RA.stream,SS_read_act);
	    /* Should user to have some way to interrupt progress ? 
	       Cancelable version is smtp_command_reset_c()
	     */
	}
	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset: ... wait ended\n"));
    }

    smtp_command_reset_tail(I);
    
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_reset: done%s\n",
		     smtp_idle == I->state ? " (state: smtp_idle)" : ""));
}

static int smtp_command_reset_c P_((struct mailer_info_1 *I,
				    struct cancel_data *cd));
static int smtp_command_reset_c(I,cd)
     struct mailer_info_1 *I;
     struct cancel_data *cd;
{
    int status = 0;

   if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_command_reset_c",
	      "Bad magic number",0);

    DPRINT(Debug,12,(&Debug,
		     "smtp_command_reset_c: mailer info, I=%p\n",I));

    if (smtp_command     == I->state) {
	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset_c: writing data to server...\n"));

	while ( smtp_command  == I->state &&
		NULL         != I->RA.stream) {

	    if (!WaitStreamFor_c(I->RA.stream,SS_write_act,cd)) {
		DPRINT(Debug,7,(&Debug,"smtp_command_reset_c: wait not succeed?\n"));

		if (cd && is_canceled(cd)) {
		    DPRINT(Debug,7,(&Debug,"smtp_command_reset_c: wait canceled!\n"));

		    smtp_cancel_reset_stream(I);
		    status = 0;
		    goto clean;
		}
	    }
	}

	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset_c: ... write ended\n"));
    }

    if (smtp_wait_reply  == I->state ||
	smtp_reply_incomplete == I->state) {

	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset_c: waiting reply from server...\n"));

	while ( (smtp_wait_reply       == I->state || 
		 smtp_reply_incomplete == I->state) &&
		NULL         != I->RA.stream) {
	
	    if (!WaitStreamFor_c(I->RA.stream,SS_read_act,cd)) {
		DPRINT(Debug,7,(&Debug,"smtp_command_reset_c: wait not succeed?\n"));
		
		if (cd && is_canceled(cd)) {
		    DPRINT(Debug,7,(&Debug,"smtp_command_reset_c: wait canceled!\n"));
		    
		    smtp_cancel_reset_stream(I);
		    status = 0;
		    goto clean;
		}
	    }
	}

	DPRINT(Debug,7,(&Debug,
			"smtp_command_reset_c: ... wait ended\n"));
    }

    smtp_command_reset_tail(I);

    status = 1;
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_reset_: done%s\n",
		     smtp_idle == I->state ? " (state: smtp_idle)" : ""));

clean:
    DPRINT(Debug,12,(&Debug,
		     "smtp_command_reset_c=%d\n",status));
    return status;
}

static const struct string *smtp_Server(I)
     struct mailer_info_1 *I;
{
    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_Server",
	      "Bad magic number",0);

    return Server_s((I->SMTP_flags & SMTP_submission) ? Submission_s : SMTP_s);
}

struct smtp_callbacks SMTP_FUNCTIONS = {
    smtp_start_command,
    smtp_push_data,
    smtp_command_arg,
    smtp_command_ready,
    smtp_command_ok,
    smtp_command_response,
    smtp_command_reset,
    smtp_Server
};

static void free_ESMTP_keywords  P_((struct mailer_info_1 *I));
static void free_ESMTP_keywords(I)
     struct mailer_info_1 *I;
{
    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"free_ESMTP_keywords",
	      "Bad magic number",0);

    if (I->ESMTP_keywords) {
	int i;
	for (i = 0; i < I->ESMTP_keyword_len; i++) {
	    if (I->ESMTP_keywords[i]) {
		free(I->ESMTP_keywords[i]);
		I->ESMTP_keywords[i] = NULL;
	    }
	}
	free(I->ESMTP_keywords);
	I->ESMTP_keywords        = NULL;
    }    
    I->ESMTP_keyword_len     = -1;   /* -1 == EHLO not issued yet */

}

			       

static int smtp_run_ehlo_after_response P_((struct mailer_info_1 *I,
					    struct SMTP_ext_libs *  smtp_ext_libs,
					    int                     smtp_ext_libcount,
					    enum CAPA_phase  phase,
					    int * need_retry_ehlo));
static int smtp_run_ehlo_after_response(I,smtp_ext_libs,smtp_ext_libcount,
					phase,need_retry_ehlo)
     struct mailer_info_1  * I;
     struct SMTP_ext_libs  * smtp_ext_libs;
     int                     smtp_ext_libcount;
     enum CAPA_phase         phase;
     int                   * need_retry_ehlo;
{
    int ret = 0;

    DPRINT(Debug,13,(&Debug,
		     "%d libraries available\n",smtp_ext_libcount));

    *need_retry_ehlo = 0;

    if (!handle_smtp_ext_libs(I,&smtp_ext_libs,&smtp_ext_libcount,
			      &phase,&SMTP_FUNCTIONS)) {
	/* Something gone wrong ... */
	
	ret = 0;
	goto clean;	
    }

    switch(phase) {
	
    case capa_prelogin_again:
	*need_retry_ehlo = 1;
	break;
	
    case capa_logged_again:
	I->SMTP_flags |= SMTP_logged;
	*need_retry_ehlo = 1;
	break;
	    
    case capa_logged:
	I->SMTP_flags |= SMTP_logged;
	break;
    default:
	break;
    }

    ret = 1;
		
 clean:
    return ret;
}

static void smtp_run_ehlo_handle_line P_((struct mailer_info_1 *I,
					  struct SMTP_ext_libs ** smtp_ext_libs,
					  int                   * smtp_ext_libcount,
					  const char * line));
static void smtp_run_ehlo_handle_line(I,smtp_ext_libs,smtp_ext_libcount,line)
     struct mailer_info_1 *I;
     struct SMTP_ext_libs ** smtp_ext_libs;
     int                   * smtp_ext_libcount;
     const char            * line;
{
    char * ext, *arg = NULL;

    I->ESMTP_keywords = 
	safe_array_realloc(I->ESMTP_keywords,
			   (I->ESMTP_keyword_len+1),
			   sizeof (*(I->ESMTP_keywords)));
    I->ESMTP_keywords[I->ESMTP_keyword_len++] =
	safe_strdup(line);	

    ext = safe_strdup(line);
    if (NULL != (arg = strchr(ext,' '))) {
	*arg = '\0';
	arg++;
    }
    
    DPRINT(Debug,2,(&Debug, 
		    "SMTP server extension: '%s'%s%s\n",
		    ext,
		    arg ? " args: " : "",
		    arg ? arg : ""));

    probe_smtp_ext_lib(smtp_ext_libs,
		       smtp_ext_libcount,
		       ext,arg);
			
    free(ext);
}

static int smtp_run_ehlo P_((struct mailer_info_1 *I));
static int smtp_run_ehlo(I)
     struct mailer_info_1 *I;
{
    int ret = 1;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_run_ehlo",
	      "Bad magic number",0);

    DPRINT(Debug,14,(&Debug,
		     "smtp_run_ehlo: mailer info, I=%p\n",I));

 retry_ehlo:
    if (smtp_start_command(I,"EHLO")) {
	char response_code[4];

	smtp_command_arg(I,hostfullname);
	smtp_command_ready(I,0);

	if (smtp_command_ok(I,response_code)) {

	    if (0 == memcmp(response_code,"250",3)) {
		int i;
		const char * line;
		
		struct SMTP_ext_libs *  smtp_ext_libs = NULL;
		int                     smtp_ext_libcount = 0;

		enum CAPA_phase  phase = capa_prelogin;
		int need_retry_ehlo = 0;

		if (I->SMTP_flags & SMTP_logged) {
		    phase = capa_logged;
		    DPRINT(Debug,13,(&Debug,"Already logged in\n"));
		}

		free_ESMTP_keywords(I);

		I->ESMTP_keyword_len = 0;

		for (i = 0; NULL != (line = smtp_command_response(I,&i)); ) {
		    if (i > 0) { /* First response line is greeting */
			smtp_run_ehlo_handle_line(I,&smtp_ext_libs,
						  &smtp_ext_libcount,
						  line);			
		    }
		}

		if (smtp_run_ehlo_after_response(I,smtp_ext_libs,
						 smtp_ext_libcount,
						 phase,&need_retry_ehlo)) {

		    if (need_retry_ehlo) {
			DPRINT(Debug,14,(&Debug,
					 "smtp_run_ehlo: post process retries EHLO\n"));
			goto retry_ehlo;
		    }

		} else {
		    DPRINT(Debug,14,(&Debug,
				     "smtp_run_ehlo: post process failed\n"));
		    ret = 0;
		    goto clean;	
		}
	       
	    } else {
		DPRINT(Debug,14,(&Debug,
				 "smtp_run_ehlo: not expected response code: %.3s\n",
				 response_code));
		goto fail;
	    }
	} else {
	fail:
	    /* If EHLO fails try HELO ... (ignoring result) */
	    smtp_command_reset(I);
	    if (smtp_start_command(I,"HELO")) {
		smtp_command_arg(I,hostfullname);
		smtp_command_ready(I,0);

		I->ESMTP_keyword_len = 0;
	    } else 
		goto failure;
	}

    } else {
    failure:
	DPRINT(Debug,14,(&Debug,
			 "smtp_run_ehlo: failure\n"));
	ret = 0;

	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpConnectionFailed,
			  "Connection to %S server failed"),
		  Server_s((I->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s));

    }

 clean:
    smtp_command_reset(I);

    DPRINT(Debug,14,(&Debug,
		     "smtp_run_ehlo=%d: done\n",ret));

    return ret;
}

static int smtp_run_ehlo_c P_((struct mailer_info_1 *I,struct cancel_data *cd));
static int smtp_run_ehlo_c(I,cd)
     struct mailer_info_1 *I;
     struct cancel_data *cd;
{
    int ret = 1;
    
    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_run_ehlo_c",
	      "Bad magic number",0);

    DPRINT(Debug,14,(&Debug,
		     "smtp_run_ehlo_c: mailer info, I=%p\n",I));

 retry_ehlo:
    if (smtp_start_command_c(I,"EHLO",cd)) {
	char response_code[4];
	
	smtp_command_arg(I,hostfullname);	
	if (smtp_command_ready_c(I,0,cd)) {
	    DPRINT(Debug,7,(&Debug,"smtp_run_ehlo_c: EHLO done\n"));
	} else
	    goto failure;
      
	if (smtp_command_ok_c(I,response_code,cd)) {

	    if (0 == memcmp(response_code,"250",3)) {
		int i;
		const char * line;
		
		struct SMTP_ext_libs *  smtp_ext_libs = NULL;
		int                     smtp_ext_libcount = 0;

		enum CAPA_phase  phase = capa_prelogin;
		int need_retry_ehlo = 0;

		if (I->SMTP_flags & SMTP_logged) {
		    phase = capa_logged;
		    DPRINT(Debug,13,(&Debug,"Already logged in\n"));
		}

		free_ESMTP_keywords(I);

		I->ESMTP_keyword_len = 0;

		for (i = 0; NULL != (line = smtp_command_response_c(I,&i,cd)); ) {
		    if (i > 0) { /* First response line is greeting */
			smtp_run_ehlo_handle_line(I,&smtp_ext_libs,
						  &smtp_ext_libcount,
						  line);
		    }
		}

		if (cd && is_canceled(cd)) {
		    DPRINT(Debug,7,(&Debug,"smtp_run_ehlo_c: command canceled!\n"));
		    
		    ret = 0;
		    goto clean;
		}

		if (smtp_run_ehlo_after_response(I,smtp_ext_libs,
						 smtp_ext_libcount,
						 phase,&need_retry_ehlo)) {

		    if (need_retry_ehlo) {
			DPRINT(Debug,14,(&Debug,
					 "smtp_run_ehlo_c: post process retries EHLO\n"));
			goto retry_ehlo;
		    }

		} else {
		    DPRINT(Debug,14,(&Debug,
				     "smtp_run_ehlo_c: post process failed\n"));
		    ret = 0;
		    goto clean;	
		}

	    } else {
		DPRINT(Debug,14,(&Debug,
				 "smtp_run_ehlo_c: not expected response code: %.3s\n",
				 response_code));
		goto fail;
	    }

	} else {
	fail:

	    if (cd && is_canceled(cd)) {
		DPRINT(Debug,7,(&Debug,"smtp_run_ehlo_c: command canceled!\n"));
		
		ret = 0;
		goto clean;
	    }

	    if (smtp_command_reset_c(I,cd)) {
		if (smtp_start_command_c(I,"HELO",cd)) {
		    smtp_command_arg(I,hostfullname);

		    if (smtp_command_ready_c(I,0,cd)) {
			DPRINT(Debug,7,(&Debug,"smtp_run_ehlo_c: HELO done\n"));
		    } else 
			goto failure;
		    
		} else 
		    goto failure;
	    } else
		goto failure;
	}

    } else {
    failure:
	DPRINT(Debug,14,(&Debug,
			 "smtp_run_ehlo_c: failure\n"));
	ret = 0;

	if (cd && is_canceled(cd)) {
	    DPRINT(Debug,7,(&Debug,"smtp_run_ehlo_c: command canceled!\n"));
	    
	    goto clean;
	}

	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpConnectionFailed,
			  "Connection to %S server failed"),
		  Server_s((I->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s));

    }
   
 clean: 
    if (smtp_command_reset_c(I,cd)) {
	if (ret) {
	    DPRINT(Debug,14,(&Debug,
			     "smtp_run_ehlo_c: done\n"));
	}
    } else {
	DPRINT(Debug,14,(&Debug,
			 "smtp_run_ehlo_c: Command reset failed\n"));
	ret = 0;
    }

    DPRINT(Debug,14,(&Debug,
		     "smtp_run_ehlo_c=%d\n",ret));

    return ret;
}

/* common to submit and sendmail-bs mailer -------------------------------- */


S_(mailer_set_option default_set_option)
static int default_set_option  P_((struct mailer_config *M,
				   struct mailer_option_list *X,
				   char *value));
static int default_set_option(M,X,value) 
     struct mailer_config *M;
     struct mailer_option_list *X;
     char *value;
{
    int r ;

    r = X->value_func(M,X,&value,1);
    return r;
}

static char * default_from P_((void));
static char * default_from(void)
{
    char * ret = NULL;

    const char * maildomain = hostfullname;

    const char **mailname_list = give_dt_path_as_elems(&mailname,"mailname");

    if (mailname_list && mailname_list[0] && mailname_list[0][0])
	maildomain = mailname_list[0];
    else {
	DPRINT(Debug,10,(&Debug,
			 "default_from: mailname not set, using hostfullname=%s\n",maildomain));
    }
	

    if (NULL != strpbrk(username,
                        ":\"\\!#%&()=?',;.:<>")) {
	ret = elm_message(FRM("<%Q@%s>"),
			   username,maildomain);
    } else {
	ret = elm_message(FRM("<%s@%s>"),
			   username,maildomain);
    }

    DPRINT(Debug,13,(&Debug,
		     "default_from=%s\n",ret));

    return ret;
}

static enum msd_addr_status msd_addr_response  P_((char response_code[4]));
static enum msd_addr_status msd_addr_response(response_code)
     char response_code[4];
{
    switch (response_code[0]) {
    case '4':
	return  msd_addr_tmpfail;
    case '5':
	return msd_addr_fail;
    case '2':
	return msd_addr_ok;
    }

    return msd_addrstat_none;  /* Bad response code */
}

static enum msd_data_status msd_data_response  P_((char response_code[4]));
static enum msd_data_status msd_data_response(response_code)
     char response_code[4];
{
    switch (response_code[0]) {
    case '4':
	return msd_mail_tmpfail;
    case '5':
	return msd_mail_fail;
    case '3':
	return msd_data_ok;
    case '2':
	return msd_mail_sent;
    }

    return msd_datastat_none;  /* Bad response code */
}

static void translate_response P_((char response_code[4], int *Exit_stat));
static void translate_response(response_code,Exit_stat)
     char response_code[4];
     int *Exit_stat;
{

#ifdef I_SYSEXITS
    switch (response_code[0]) {
    case '4':
#ifdef  EX_TEMPFAIL
	*Exit_stat = EX_TEMPFAIL;
#endif
	break;
    case '5':
#ifdef  EX_UNAVAILABLE 
	*Exit_stat = EX_UNAVAILABLE;
#endif
	break;
    }

#ifdef  EX_NOUSER   		
    if (0 == strncmp(response_code,"550",3))
	*Exit_stat = EX_NOUSER;   /* Can this to be used for sender? */
#endif
#endif		

}

/* Return 1 if  if mail is moved to ticket */
static int smtp_pass_ticket P_((struct mailer_config *M,
				struct mail_send_state *C,
				struct mailer_cancel_ticket * ticket,
				enum msd_action action));
static int smtp_pass_ticket(M,C,ticket,action)
     struct mailer_config *M;
     struct mail_send_state *C;
     struct mailer_cancel_ticket * ticket;
     enum msd_action action;
{
    int r = 0;

    if (MAILER_CANCEL_TICKET_magic != ticket->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_pass_ticket",
	      "Bad magic number (mailer_cancel_ticket)",0);

    if (C->magic != MAILER_magic)
    	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_pass_ticket",
	      "Bad magic number (mail_send_state)",0);

    ticket->msd_action = action;

    switch (ticket->msd_action) {
    case msd_act_EOF: 
	DPRINT(Debug,4, (&Debug, 
			 "smtp_pass_ticket: action msd_act_EOF\n"));
	break;
	
    case msd_act_SIG:
	DPRINT(Debug,4, (&Debug, 
			 "smtp_pass_ticket: action msd_act_SIG\n"));
	break;
	
    case msd_act_need_restart:
	DPRINT(Debug,4, (&Debug, 
			 "smtp_pass_ticket: action msd_act_need_restart\n"));
	r = 1;
	break;
	
    case msd_act_default:
	DPRINT(Debug,4, (&Debug, 
			 "smtp_pass_ticket: action msd_act_default\n"));
	break;
	
    case msd_act_cancel:
	DPRINT(Debug,4, (&Debug, 
			 "smtp_pass_ticket action msd_act_cancel\n"));
	r = 1;
	break;
	
    case msd_act_edit:
	DPRINT(Debug,4, (&Debug, 
			 "smtp_pass_ticket: action msd_act_edit\n"));
	r = 1;
	break;
	
    case msd_act_send:
	DPRINT(Debug,4, (&Debug, 
			 "smtp_pass_ticket: action msd_act_send\n"));
	break;
    default:
	DPRINT(Debug,4, (&Debug, 
			 "smtp_pass_ticket: action %d\n",
			 action));
    }

    if (r) {
	
	ticket->fname = C->fname;
	C->fname = NULL;
	
	ticket->OUT   = C->OUT;
	C->OUT        = NULL;
	
	if (C->OUT) 
	    out_state_fseek (C->OUT, 0);
	
	if (! ticket->fname) {
	    DPRINT(Debug,4, (&Debug, 
			     "smtp_pass_ticket: No filename\n"));
	}
	
	if (! ticket->OUT) {
	    DPRINT(Debug,4, (&Debug, 
			     "smtp_pass_ticket: No out_state\n"));
	}
	
	if (! ticket->fname && ! ticket->OUT)   
	    r = 0;
    }

    DPRINT(Debug,8, (&Debug, 
		     "smtp_pass_ticket=%d\n",r));
	
    return r;
    
}

static void smtp_print_response_tail P_((struct mailer_info_1 *I,
					 int *idx));
static void smtp_print_response_tail(I,idx)
     struct mailer_info_1 *I;
     int *idx;
{
    const char *r;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_print_response_tail",
	      "Bad magic number",0);
    
    while (NULL != (r = smtp_command_response(I,idx))) {
	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpServerMsg,
			  "%S server: %s"),
		  smtp_Server(I),
		  r);
    }    
}

static struct string * message_helper P_((struct mailer_info_1 *I,
					  char response_code[4]));
static struct string * message_helper(I,response_code)
     struct mailer_info_1 *I;
     char response_code[4];
{
    int idx = 0;
    const char * r;
    struct string * message;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"message_helper",
	      "Bad magic number",0);

    r = smtp_command_response(I,&idx);

    if (r)
	message = format_string(FRM("%3.3s %s"),
				response_code,r);
    else
	message = format_string(FRM("%3.3s"),
				response_code);

    return message;
}

S_(mailer_backend3_hook smtp_common_mailer_backend3)
static int smtp_common_mailer_backend3 P_((struct mailer_config *M,
					   struct mail_send_state **C,
					   int encoding_top,
					   char * title,
					   sending_message_func *sm,
					   int *exit_status,
					   struct mailer_cancel_ticket * 
					   ticket,
					   long message_size));
static int smtp_common_mailer_backend3(M,C,encoding_top,title,sm,
				       exit_status,ticket,message_size)
     struct mailer_config *M;
     struct mail_send_state **C;
     int encoding_top;
     char * title;
     sending_message_func *sm;
     int *exit_status;   
     struct mailer_cancel_ticket * ticket;
     long message_size;
{
    char buffer [ 32 * 1024 ];
    int l;
    int have_dsn = 0;
    int have_size = 0;
    int Exit_stat = -1;
    int ret = 0; 

    char * from;
    char * f1;
    
    int fail = 0; /* Number of failed addresses */

    struct string * from_message    =  NULL;
    struct string * data_message    =  NULL;
    enum msd_action          action = msd_act_default;

    if ((*C)->magic != MAILER_magic)
    	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_common_mailer_backend3",
	      "Bad magic number (mail send state)",0);

    f1 = (*C)->mail_from;
    if (!f1)
	f1 = default_from();

    if ('<' == f1[0]) {
	from = elm_message(FRM("FROM:%s"),
			   f1);
    } else {
	from = elm_message(FRM("FROM:<%s>"),
			   f1);
    }
    
    if ((*C)->msd_sd) 
	from_message = new_string2(system_charset,s2us(f1));
	    
    if (f1 != (*C)->mail_from) {
	free(f1);
	f1 = NULL;
    }

    if (SMTP_mailer_info_magic != (*C)->head->p.p1->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_common_mailer_backend",
	      "Bad magic number (smtp mailer info)",0);

    if ((*C)->msd_sd && 
	(*C)->head->p.p1->RA.host) {
	
	struct string * message = 
	    new_string2(system_charset,
			s2us((*C)->head->p.p1->RA.host));

	(*C)->msd_con_upd_cb((*C)->msd_sd,msd_server,message);

	free_string(&message);
    }

    /* With b)ounce command EHLO/HELO is not issued before that */
    if (-1 == (*C)->head->p.p1->ESMTP_keyword_len) {
	if (!smtp_run_ehlo((*C)->head->p.p1)) {
	    ret = 0; /* Some extension failed hard */
	    goto fail;
	}
    }

    if ((*C)->msd_sd && (*C)->head->p.p1->RA.stream) {
	int r = 0;
	struct string * peer_name = NULL;
	StreamInfo((*C)->head->p.p1->RA.stream,SS_verify,&r,NULL,
		   &peer_name);

	if (r > 0 && peer_name) 
	    (*C)->msd_con_upd_cb((*C)->msd_sd,msd_verified_as,
				 peer_name);

	if (peer_name)
	    free_string(&peer_name);
    }

    if ((*C)->dsn) {
	int i;
	for (i = 0; i < (*C)->head->p.p1->ESMTP_keyword_len; i++) {
	    if (0 == istrcmp("DSN",(*C)->head->p.p1->ESMTP_keywords[i])) {
		have_dsn = 1;
	    } 
	}
    }
       
    
    if (message_size >= 0) {
	int i;

	if ((*C)->msd_sd)
	    (*C)->msd_size_upd_cb((*C)->msd_sd,msd_message_size,
				   message_size);


	for (i = 0; i < (*C)->head->p.p1->ESMTP_keyword_len; i++) {
	    if (0 == istrcmp("SIZE",(*C)->head->p.p1->ESMTP_keywords[i])) {
		have_size = 1;
	    } else if (0 == strincmp("SIZE ",
				    (*C)->head->p.p1->ESMTP_keywords[i],
				    5)) {
		char * end = NULL;
		long limit = strtol((*C)->head->p.p1->ESMTP_keywords[i]+5,
				    &end,10);
		have_size = 1;

		if (end && '\0' == *end) {
		    
		    if ((*C)->msd_sd)
			(*C)->msd_size_upd_cb((*C)->msd_sd,
					      msd_size_limit,
					      limit);		    
		    
		    if (message_size > limit) {
			
			DPRINT(Debug,4,(&Debug, 
					"Message too big: size = %ld > limit = %ld\n",
					message_size,limit));
			
			/* prompt user */
			
			if ((*C)->msd_sd && ticket) {
			    
			    action = 
				(*C)->msd_size_check_cb((*C)->msd_sd,
							message_size,	    
							limit);
			    
			    if (smtp_pass_ticket(M,*C,ticket,action)) {
				DPRINT(Debug,4,(&Debug, " -- passed mail to ticket\n"));
				goto cancel_action;
			    } else { 
				switch (action) {
				case msd_act_EOF:
				case msd_act_SIG:
				case msd_act_need_restart:
				case msd_act_cancel:
				case msd_act_edit:
				    
				    DPRINT(Debug,4,(&Debug, " -- cancel request -- no ticket\n"));
				    goto cancel_action;
				case msd_act_default:
				case msd_act_send:
				    break;
				}			    
			    }				    
			} 
		    }		    
		}
	    }
	}
    }

    if (smtp_start_command((*C)->head->p.p1,"MAIL")) {
	char response_code[4];
	
	if ((*C)->msd_sd)
	    (*C)->msd_addr_upd_cb((*C)->msd_sd,-1,from_message,
				  msd_sending_addr,NULL);

	smtp_command_arg((*C)->head->p.p1,from);

	if (have_size) {
	    char * arg = elm_message(FRM("SIZE=%ld"),
				     message_size);

	    smtp_command_arg((*C)->head->p.p1,arg);
	    free(arg);
	}


	if (encoding_top == ENCODING_8BIT) {
	    int i;
	    for (i = 0; i < (*C)->head->p.p1->ESMTP_keyword_len; i++)
		if (0 == istrcmp("8BITMIME",
				 (*C)->head->p.p1->ESMTP_keywords[i])) {
		    smtp_command_arg((*C)->head->p.p1,"BODY=8BITMIME");
		    break;
		}
	}

#if 0
	if (encoding_top == ENCODING_BINARY) {
	    int i;
	    for (i = 0; i < (*C)->head->p.p1->ESMTP_keyword_len; i++)
		if (0 == istrcmp("BINARYMIME",
				 (*C)->head->p.p1->ESMTP_keywords[i])) {
		    smtp_command_arg((*C)->head->p.p1,"BODY=BINARYMIME");
		    break;
		}
	}
#endif

	if (have_dsn) {
	    if ((*C)->dsn & DSN_FULL) 
		smtp_command_arg((*C)->head->p.p1,"RET=FULL");
	    else if ((*C)->dsn & DSN_HDRS)
		smtp_command_arg((*C)->head->p.p1,"RET=HDRS");    
	}

	smtp_command_ready((*C)->head->p.p1,0);
	if (smtp_command_ok((*C)->head->p.p1,response_code)) {	    

	    if ((*C)->msd_sd) {

		struct string * message = 
		    message_helper((*C)->head->p.p1,response_code);

		(*C)->msd_addr_upd_cb((*C)->msd_sd,-1,
				      from_message,
				      msd_addr_response(response_code),
				      message);

		free_string(&message);
	    
	    }

	} else { /* command failed */

	    if (response_code[0]) {
		int idx = 0;
		const char * r = 
		    smtp_command_response((*C)->head->p.p1,&idx);

		struct string * message = NULL;
		
		if (r) {
		    lib_error(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpRejectedMail,
				      "%S server rejected MAIL with %3.3s %s"),
			      smtp_Server((*C)->head->p.p1),
			      response_code,r);
		    
		    if ((*C)->msd_sd) 
			message = format_string(FRM("%3.3s %s"),
						response_code,r);

		    smtp_print_response_tail((*C)->head->p.p1,&idx);

		} else {
		    lib_error(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpRejectedMail0,
				      "%S server rejected MAIL with code %3.3s"),
			      smtp_Server((*C)->head->p.p1),
			      response_code);

		    if ((*C)->msd_sd) 
			message = format_string(FRM("%3.3s"),
						response_code);		   
		}
		
		if ((*C)->msd_sd) {
		    if (ticket) 
			action = 
			    (*C)->msd_addr_check_cb((*C)->msd_sd,-1,
						    from_message,
					      msd_addr_response(response_code),
						    message,
						    msd_mail_failed);
		    else
			(*C)->msd_addr_upd_cb((*C)->msd_sd,-1,
					      from_message,
					      msd_addr_response(response_code),
					      message);
		}		

		if (message)
		    free_string(&message);

		translate_response(response_code,&Exit_stat);
	    } else {
		lib_error(CATGETS(elm_msg_cat, SmtpSet,
				  SmtpConnectionFailMail,
				  "Connection to %S server failed after MAIL"),
			  Server_s(((*C)->head->p.p1->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s));
		
		smtp_command_reset((*C)->head->p.p1);

		goto mail_connection_failure;
	    }

	    if (ticket && 
		smtp_pass_ticket(M,*C,ticket,action)) {
		DPRINT(Debug,4,(&Debug, " -- passed mail to ticket\n"));
	    }

	    free(from); from = NULL;
	
	    smtp_command_reset((*C)->head->p.p1);
	    goto cancel_action;
	} /* end command failed */

    } else {

	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpConnectionFailed,
			  "Connection to %S server failed"),
		  Server_s(((*C)->head->p.p1->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s));

	mail_connection_failure:
#ifdef  EX_TEMPFAIL
	Exit_stat = EX_TEMPFAIL;
#endif
	
	if ((*C)->msd_sd) {
	    /* Not actually data phase */
	    if (ticket)  {
		action = 
		    (*C)->msd_fail_check_cb((*C)->msd_sd,
					    msd_mail_tmpfail,
					    NULL,
					    msd_mail_failed);
			
		if (msd_act_send == action) {
		    /* send after reconnect */
		    
		    DPRINT(Debug,4, (&Debug,
				     "  >> msd_act_send >> msd_act_need_restart\n"));

		    action = msd_act_need_restart;
		}
	    } else
		(*C)->msd_data_upd_cb((*C)->msd_sd,msd_mail_tmpfail,
				      NULL);
	}

	if (ticket && 
	    smtp_pass_ticket(M,*C,ticket,action)) {
	    DPRINT(Debug,4,(&Debug, " -- passed mail to ticket\n"));
	}

	free(from); from = NULL;

	goto cancel_action;
    }
    smtp_command_reset((*C)->head->p.p1);

    free(from); from = NULL;

    if ((*C)->addrs) {
	int i;

	for (i = 0; (*C)->addrs[i]; i++) {
	    char * R = (*C)->addrs[i];

	    struct string * rcpt_message = NULL;

	    if ((*C)->msd_sd)
		rcpt_message = new_string2(system_charset,
					   s2us((*C)->addrs[i]));

	    /* Make sure that addresses are inside of <> */
	    if ('<' != (*C)->addrs[i][0])
		R = kludge_addr( &((*C)->addrs[i]));

	    if (smtp_start_command((*C)->head->p.p1,"RCPT")) {
		char response_code[4];
		char * rcpt = safe_strdup("TO:");

		if ((*C)->msd_sd)
		    (*C)->msd_addr_upd_cb((*C)->msd_sd,i,rcpt_message,
					  msd_sending_addr,NULL);

		rcpt = strmcat(rcpt,R);
		smtp_command_arg((*C)->head->p.p1,rcpt);
		free(rcpt); rcpt = NULL;

		if (have_dsn) {
		    if ((*C)->dsn & DSN_NEVER)
			smtp_command_arg((*C)->head->p.p1,"NOTIFY=NEVER");
		    
		    else if ((*C)->dsn & (DSN_SUCCESS|DSN_FAILURE|DSN_DELAY)) {
			char * A = safe_strdup("NOTIFY=");
			int x = 0;
			if ((*C)->dsn & DSN_SUCCESS) {
			    A = strmcat(A,"SUCCESS"); x++;
			}
			if ((*C)->dsn & DSN_FAILURE) {
			    if (x) A = strmcat(A,",");
			    A = strmcat(A,"FAILURE"); x++;
			}
			if ((*C)->dsn & DSN_DELAY) {
			    if (x) A = strmcat(A,",");
			    A = strmcat(A,"DELAY"); 
			}
			smtp_command_arg((*C)->head->p.p1,A);
			free(A); A = NULL;
		    }
		}

		smtp_command_ready((*C)->head->p.p1,0);
		if (smtp_command_ok((*C)->head->p.p1,response_code)) {

		    if ((*C)->msd_sd) {

			struct string * message = 
			    message_helper((*C)->head->p.p1,response_code);
			
			(*C)->msd_addr_upd_cb((*C)->msd_sd,i,
					      rcpt_message,
					      msd_addr_response(response_code),
					      message);

			free_string(&message);
			
		    }

		} else { /* command failed */

		    if (response_code[0]) {
			int idx = 0;
			const char * r = 
			    smtp_command_response((*C)->head->p.p1,&idx);
			
			struct string * message = NULL;

			if (r) {
			    lib_error(CATGETS(elm_msg_cat, SmtpSet,
					      SmtpRejectedCommand,
					      "%S server rejected %s with %3.3s %s"),
				      smtp_Server((*C)->head->p.p1),
				      (*C)->addrs[i],
				      response_code,r);

			    if ((*C)->msd_sd) 
				message = format_string(FRM("%3.3s %s"),
							response_code,r);

			    smtp_print_response_tail((*C)->head->p.p1,&idx);
			} else {
			    lib_error(CATGETS(elm_msg_cat, SmtpSet,
					      SmtpRejectedCommand0,
					      
					      "%S server rejected %s with code %3.3s"),
				      smtp_Server((*C)->head->p.p1),
				      (*C)->addrs[i],
				      response_code);
			    if ((*C)->msd_sd) 
				message = format_string(FRM("%3.3s"),
							response_code);	
			}

			if ((*C)->msd_sd) {
			    if (ticket) {
				enum msd_can_continue can_continue = 
				    msd_address_problem;
				
				if (i == (*C)->addr_count-1 &&
				    fail >= i) {
				    
				    DPRINT(Debug,8,(&Debug, 
						    " -- last address failed"));
				    if (i) {
					DPRINT(Debug,8,(&Debug, 
							"and also previous (%d) addresses are failed",
							fail));
				    }
				    
				    DPRINT(Debug,8,(&Debug,", %d addresses\n",(*C)->addr_count));

				    can_continue = msd_mail_failed;
				}

				action = 
				    (*C)->msd_addr_check_cb((*C)->msd_sd,i,
							    rcpt_message,
					     msd_addr_response(response_code),
							    message,
							    can_continue);
			    } else
				(*C)->msd_addr_upd_cb((*C)->msd_sd,i,
						      rcpt_message,
					     msd_addr_response(response_code),
						      message);
			}

			if (message)
			    free_string(&message);

			translate_response(response_code,&Exit_stat);
		    } else {
			lib_error(CATGETS(elm_msg_cat, SmtpSet,
					  SmtpConnectionFailRCPT,
					  "Connection to %S server failed after RCPT"),
				  Server_s(((*C)->head->p.p1->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s));
			
			smtp_command_reset((*C)->head->p.p1);			
			goto rcpt_connection_failure;
		    }
		    fail++;

		    if (ticket && 
			smtp_pass_ticket(M,*C,ticket,action)) {
			DPRINT(Debug,4,(&Debug, " -- passed mail to ticket\n"));

			if (rcpt_message)
			    free_string(&rcpt_message);

			smtp_command_reset((*C)->head->p.p1);
			goto cancel_action;
		    } else if ((*C)->msd_sd) { 
			switch (action) {
			case msd_act_EOF:
			case msd_act_SIG:
			case msd_act_need_restart:
			case msd_act_cancel:
			case msd_act_edit:
			    
			    DPRINT(Debug,4,(&Debug, " -- cancel request -- no ticket\n"));

			    if (rcpt_message)
				free_string(&rcpt_message);

			    smtp_command_reset((*C)->head->p.p1);
			    goto cancel_action;
			case msd_act_default:
			case msd_act_send:
			    break;
			}
		    }
		    	
		} /* end command failed */

	    } else {
		lib_error(CATGETS(elm_msg_cat, SmtpSet,
				  SmtpConnectionFailed,
				  "Connection to %S server failed"),
			  Server_s(((*C)->head->p.p1->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s));

	    rcpt_connection_failure:
		fail++;

#ifdef  EX_TEMPFAIL
		Exit_stat = EX_TEMPFAIL;
#endif
	
		if ((*C)->msd_sd) {
		    /* Not actually data phase */
		    if (ticket)  {
			action = 
			    (*C)->msd_fail_check_cb((*C)->msd_sd,
						    msd_mail_tmpfail,
						    NULL,
						    msd_mail_failed);
			
			if (msd_act_send == action) {
			    /* send after reconnect */
			    
			    DPRINT(Debug,4, (&Debug,
					     "  >> msd_act_send >> msd_act_need_restart\n"));
			    
			    action = msd_act_need_restart;
			}
		    } else
			(*C)->msd_data_upd_cb((*C)->msd_sd,msd_mail_tmpfail,
					      NULL);
		}
	
		if (ticket && 
		    smtp_pass_ticket(M,*C,ticket,action)) {
		    DPRINT(Debug,4,(&Debug, " -- passed mail to ticket\n"));
		}
	
		if (rcpt_message)
		    free_string(&rcpt_message);
		
		free_mail_send_state(C);
		ret = 0; /* FAIL */
		goto fail;       
	    }
	    if (rcpt_message)
		free_string(&rcpt_message);

	    smtp_command_reset((*C)->head->p.p1);
	}

	if (fail) {
	    lib_error(CATGETS(elm_msg_cat, SmtpSet,
			      SmtpRcptsFailed,
			      "%d recipients from %d failed"),
		      fail,i);

	    if ((*C)->msd_sd)  {
		enum msd_data_status  mail_status =
		    msd_datastat_none;

		if (fail >= i)
		    mail_status = msd_mail_fail;

		data_message = 
		    format_string(CATGETS(elm_msg_cat, SmtpSet,
					  SmtpRcptsFailed,
					  "%d recipients from %d failed"),
				  fail,i);
		
		if (ticket) {
		    
		    enum msd_can_continue can_continue = 
			msd_address_problem;
		    
		    if (fail >= i) {
			DPRINT(Debug,8,(&Debug, 
					" .. all (%d) addresses are failed, %d addresses\n",
					fail,(*C)->addr_count));
			can_continue = msd_mail_failed;
		    }
		    
		    action = 
			(*C)->msd_fail_check_cb((*C)->msd_sd,
						mail_status,
						data_message,
						can_continue);
		} else
		    (*C)->msd_data_upd_cb((*C)->msd_sd,
					  mail_status,
					  data_message);    
	    }

	    if (ticket && 
		smtp_pass_ticket(M,*C,ticket,action)) {
		DPRINT(Debug,4,(&Debug, " -- passed mail to ticket\n"));

		goto cancel_action;
	    } else if ((*C)->msd_sd) { 
		switch (action) {
		case msd_act_EOF:
		case msd_act_SIG:
		case msd_act_need_restart:
		case msd_act_cancel:
		case msd_act_edit:
		    
		    DPRINT(Debug,4,(&Debug, " -- cancel request -- no ticket\n"));		    		    
		    goto cancel_action;
		case msd_act_default:
		case msd_act_send:
		    break;
		}
	    }

	    if (fail >= i) 	       
		goto cancel_action;
	}

    } else {
	DPRINT(Debug,13,(&Debug,
			 "smtp_common_mailer_backend: No recipient addresses...\n"));
	
	goto cancel_action;
    }


    if ((*C)->msd_sd) 
	(*C)->msd_data_upd_cb((*C)->msd_sd,msd_sending_data,
			      data_message);
	
    if (smtp_start_command((*C)->head->p.p1,"DATA")) {
	char response_code[4];

	if ((*C)->msd_sd)
	    (*C)->msd_data_upd_cb((*C)->msd_sd,
				  msd_datastat_none,
				  data_message);  	    	    
       	
	smtp_command_ready((*C)->head->p.p1,0);

	if (smtp_command_ok((*C)->head->p.p1,response_code)) {

	    if ((*C)->msd_sd) {
		struct string * message =
		    data_message ?
		    data_message :
		    message_helper((*C)->head->p.p1,response_code);

		(*C)->msd_data_upd_cb((*C)->msd_sd,
				      msd_data_response(response_code),
				      message);

		if (message != data_message)
		    free_string(&message);

	    }

	} else { /* command failed */
	    	    
	    if (response_code[0]) {

		int idx = 0;
		const char * r = 
		    smtp_command_response((*C)->head->p.p1,&idx);
		
		struct string * message = NULL;

		if (r) {
		    lib_error(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpRejectedData,
				      "%S server rejected DATA with %3.3s %s"),
			      smtp_Server((*C)->head->p.p1),
			      response_code,r);

		    if ((*C)->msd_sd) 
			message = format_string(FRM("%3.3s %s"),
						response_code,r);

		    smtp_print_response_tail((*C)->head->p.p1,&idx);
		} else {
		    lib_error(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpRejectedData0,
				      "%s server rejected DATA with code %3.3s"),
			      smtp_Server((*C)->head->p.p1),
			      response_code);

		    if ((*C)->msd_sd) 
			message = format_string(FRM("%3.3s"),
						response_code);	
		    
		}

		if ((*C)->msd_sd) {
		    if (ticket) 
			action = 
			    (*C)->msd_fail_check_cb((*C)->msd_sd,
						    msd_data_response(response_code),
						    message,
						    msd_mail_failed);
		    else
			(*C)->msd_data_upd_cb((*C)->msd_sd,
					      msd_data_response(response_code),
					      message);
		    
		}

		if (message)
		    free_string(&message);

		translate_response(response_code,&Exit_stat);

	    } else {
		lib_error(CATGETS(elm_msg_cat, SmtpSet,
				  SmtpConnectionFailData,
				  "Connection to %S server failed after DATA"),
			  Server_s(((*C)->head->p.p1->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s));
		
		smtp_command_reset((*C)->head->p.p1);

		goto data_connection_failure;		
	    }
	    
	    if (ticket && 
		smtp_pass_ticket(M,*C,ticket,action)) {
		DPRINT(Debug,4,(&Debug, " -- passed mail to ticket\n"));
	    }

	    goto cancel_action;
	} /* end command failed */

    } else {

	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpConnectionFailed,
			  "Connection to %S server failed"),
		  smtp_Server((*C)->head->p.p1));

    data_connection_failure:
	
#ifdef  EX_TEMPFAIL
	if (-1 == Exit_stat)
	    Exit_stat = EX_TEMPFAIL;
#endif
	if ((*C)->msd_sd) {
	    if (ticket)  {
		action = 
		    (*C)->msd_fail_check_cb((*C)->msd_sd,
					    msd_mail_tmpfail,
					    data_message,
					    msd_mail_failed);
		
		if (msd_act_send == action) {
		    /* send after reconnect */
		    
		    DPRINT(Debug,4, (&Debug,
				     "  >> msd_act_send >> msd_act_need_restart\n"));
		    
		    action = msd_act_need_restart;
		}
	    } else
		(*C)->msd_data_upd_cb((*C)->msd_sd,msd_mail_tmpfail,
				      data_message);
	}
		
	if (ticket && 
	    smtp_pass_ticket(M,*C,ticket,action)) {
	    DPRINT(Debug,4,(&Debug, " -- passed mail to ticket\n"));
	}

	goto cancel_action;	
    }

    free_response_text((*C)->head->p.p1);
    out_state_fseek ((*C)->OUT, 0);

    /* may be incorrect ... 
     * (but here should not be other \r or \n than \r\n) 
     */
    
    while (0 < (l = mail_gets(buffer,sizeof buffer,
				  out_state_FILE((*C)->OUT)))) {
	
	if (!smtp_push_data((*C)->head->p.p1,buffer,l,1)) 
	    goto hard_failure;	
    }
    
    if (smtp_push_data((*C)->head->p.p1,".",1,0)) {
	char response_code[4];
	smtp_command_ready((*C)->head->p.p1,1);   /* End data */

	if (smtp_command_ok((*C)->head->p.p1,response_code)) {	    

	    if ((*C)->msd_sd) {
		struct string * message =
		    data_message ?
		    data_message :
		    message_helper((*C)->head->p.p1,response_code);

		(*C)->msd_data_upd_cb((*C)->msd_sd,
				      msd_data_response(response_code),
				      message);

		if (message != data_message)
		    free_string(&message);

	    }

	} else {  /* command failed */
	    if (response_code[0]) {
		int idx = 0;
		const char * r = 
		    smtp_command_response((*C)->head->p.p1,&idx);

		struct string * message = NULL;

		if (r) {
		    lib_error(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpRejectedMss,
				      "%S server rejected message with %3.3s %s"),
			      smtp_Server((*C)->head->p.p1),
			      response_code,r);

		    if ((*C)->msd_sd) 
			message = format_string(FRM("%3.3s %s"),
						response_code,r);

		    smtp_print_response_tail((*C)->head->p.p1,&idx);
		} else {
		    lib_error(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpRejectedMss0,
				      "%S server rejected message with code %3.3s"),
			      smtp_Server((*C)->head->p.p1),
			      response_code);
		     if ((*C)->msd_sd) 
			message = format_string(FRM("%3.3s"),
						response_code);
		}

		if ((*C)->msd_sd) {
		    if (ticket) 
			action = 
			    (*C)->msd_fail_check_cb((*C)->msd_sd,
						    msd_data_response(response_code),
						    message,
						    msd_mail_failed);
		    else
			(*C)->msd_data_upd_cb((*C)->msd_sd,
					      msd_data_response(response_code),
					      message);
		    
		}

		if (message)
		    free_string(&message);


		translate_response(response_code,&Exit_stat);	      

	    } else {

		lib_error(CATGETS(elm_msg_cat, SmtpSet,
				  SmtpConnectionFailMss,
				  "Connection to %S server failed after message"),
			  smtp_Server((*C)->head->p.p1));
				
		goto dot_connection_failure;		
	    }

	    if (ticket && 
		smtp_pass_ticket(M,*C,ticket,action)) {
		DPRINT(Debug,4,(&Debug, " -- passed mail to ticket\n"));
	    }
	    
	    goto cancel_action;
	} /* end command failed */

	{ 
	    union any_fd FD;
	    struct run_state RS;

	    FD.mail_fd         = *C;
	    RS.save_errno      = 0;

	    /* Fake ... */
	    (*C)->orig_func(FD,title,&RS,1,0);
	}
    } else {
    hard_failure:
	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpConnectionFailData2,
			  "Connection to %S server failed during DATA"),
		  smtp_Server((*C)->head->p.p1));
	
	if ((*C)->head->p.p1->RA.stream ) {
	    DPRINT(Debug,12,(&Debug,
			     "smtp_common_mailer_backend3: Closing stream\n"));
	    
	    FreeStreamStack( &((*C)->head->p.p1->RA.stream));
	}

    dot_connection_failure:

#ifdef  EX_TEMPFAIL
	if (-1 == Exit_stat)
	    Exit_stat = EX_TEMPFAIL;
#endif
	
	if ((*C)->msd_sd) {
	    if (ticket)  {
		action = 
		    (*C)->msd_fail_check_cb((*C)->msd_sd,
					    msd_mail_tmpfail,
					    data_message,
					    msd_mail_failed);
		
		if (msd_act_send == action) {
		    /* send after reconnect */
		    
		    DPRINT(Debug,4, (&Debug,
				     "  >> msd_act_send >> msd_act_need_restart\n"));
		    
		    action = msd_act_need_restart;
		}
	    } else
		(*C)->msd_data_upd_cb((*C)->msd_sd,msd_mail_tmpfail,
				      data_message);
	}				
	
	if (ticket && 
	    smtp_pass_ticket(M,*C,ticket,action)) {
	    DPRINT(Debug,4,(&Debug, " -- passed mail to ticket\n"));
	}
	
	goto cancel_action;     
    } 
       
    if (-1 == Exit_stat && 0 == fail)
	Exit_stat = 0;   /* OK */
    ret = 1;         /* OK */

    if (!ret) {
    cancel_action:
	if (smtp_start_command((*C)->head->p.p1,"RSET")) {
	    smtp_command_ready((*C)->head->p.p1,0);
	    smtp_command_reset((*C)->head->p.p1);
	}
	
	ret = 0; /* backend failed */
    }
    free_mail_send_state(C);

 fail:
    if (data_message)
	free_string(&data_message);
    if (from_message)
	free_string(&from_message);


    if (exit_status && Exit_stat >= 0)
	*exit_status = Exit_stat;

    return ret;
}



S_(mailer_init_hook smtp_common_mailer_init)
static int smtp_common_mailer_init P_((struct mailer_config *M,
				       struct mail_send_state *C,
				       struct mailer_info *I));
static int smtp_common_mailer_init(M,C,I)
     struct mailer_config *M;
     struct mail_send_state *C;
     struct mailer_info *I;
{
    FILE * F;
    const char * tmp = give_dt_estr_as_str(&temp_dir_e,"tmpdir",NULL,NULL);

    if (C->magic != MAILER_magic)
    	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_common_mailer_init",
	      "Bad magic number (mail send state)",0);

    if (!I) {
	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpMailerInfo,
			  "SMTP mailer requires info parameter"));
	return 0;
    }
    
    if (!tmp)
	return 0;

    if (SMTP_mailer_info_magic != I->p.p1->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_common_mailer_init",
	      "Bad magic number (smtp mailer info)",0);

    if (I->p.p1->SMTP_flags & SMTP_reserve) {
	panic("MAILER PANIC",__FILE__,__LINE__,
	      "smtp_common_mailer_init",
	      "SMTP transaction already on progress",0);
    }
	
    I->p.p1->SMTP_flags |= SMTP_reserve;


    if (smtp_start_command(I->p.p1,"NOOP")) {
	smtp_command_ready(I->p.p1,0);
	smtp_command_reset(I->p.p1);
    }

    
    C->fname = elm_message(FRM("%selm.sndsm-%d"), 
			   tmp, getpid());

    if (!C->fname) {
	DPRINT(Debug,1,(&Debug,   
			"couldn't make temp file nam! (mail)\n"));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCouldNotMakeTemp,
			  "Sorry - couldn't make temp file name."));
	return 0;
    }

    F = safeopen_rdwr(C->fname);

    if (!F) {
	int err = errno;
	DPRINT(Debug,1, (&Debug,
			 "Attempt to open file %s for writing failed! (write_header_info)\n",
			 C->fname));
	DPRINT(Debug,1, (&Debug, "** %s **\n\n", strerror(err)));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmErrorTryingToWrite,
			  "Error %s encountered trying to write to %s."), 
		  strerror(err), C->fname);
	
	return 0;
    }

    if (C->OUT)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_common_mailer_init",
	      "out_state already set (mail send state)",0);

    C->OUT = new_out_state(STATE_out_file);
    set_out_state_file(F,C->OUT);

    /* SMTP wants CRLF as eoln */
    set_out_state_EOLN(C->OUT,1);

    return 1;
}


S_(mailer_close_hook smtp_common_mailer_close)
static void smtp_common_mailer_close P_((struct mailer_config *M,
					 struct mail_send_state *C));
static void smtp_common_mailer_close(M,C)
     struct mailer_config *M;
     struct mail_send_state *C;
{

    if (! C->head)
	return;   /* Oh no ... */

    if (SMTP_mailer_info_magic != C->head->p.p1->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_common_mailer_close",
	      "Bad magic number (smtp mailer info)",0);

    switch (C->head->p.p1->state) {
    case smtp_need_data:
    case smtp_data:
	DPRINT(Debug,13,(&Debug,
			 "smtp_common_mailer_close: OH NO -- middle of data transfer\n"));
	C->head->p.p1->state = smtp_error;
	break;
    default:
	break;
    }

    if (smtp_start_command(C->head->p.p1,"RSET")) {
	smtp_command_ready(C->head->p.p1,0);
	smtp_command_reset(C->head->p.p1);
    }

    /* C->fname is free'ed by caller ( free_mail_send_state() ) */

    if (C->OUT) {
	FILE *F = out_state_FILE(C->OUT);
	
	free_out_state(& (C->OUT));
	if (F) {
	    fclose(F);
	}
    }

    C->head->p.p1->SMTP_flags &= ~SMTP_reserve;
}

S_(mailer_info_query_hook smtp_common_mailer_info_query)
static int smtp_common_mailer_info_query  P_((struct mailer_config *M,
					      struct mailer_info *I,
					      enum MI_query query));
static int smtp_common_mailer_info_query(M,I,query)
     struct mailer_config *M;
     struct mailer_info *I;
     enum MI_query query;
{
    if (SMTP_mailer_info_magic != I->p.p1->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_common_mailer_info_query",
	      "Bad magic number (smtp mailer info)",0);

    if (-1 == I->p.p1->ESMTP_keyword_len) {
	if (!smtp_run_ehlo(I->p.p1)) {
	    DPRINT(Debug,10,(&Debug,"Some extension failed -- ignoring now\n"));
	}
    }

    switch(query) {
	int i;

    case MI_HAVE_8BITMIME:
	for (i = 0; i < I->p.p1->ESMTP_keyword_len; i++)
	    if (0 == istrcmp("8BITMIME",I->p.p1->ESMTP_keywords[i]))
		return 1;
	DPRINT(Debug,10,(&Debug,"8BITMIME not found from EHLO response\n"));
	return 0;
    case MI_HAVE_BINARYMIME:
	return 0;       /* This code does not support BINARYMIME */
    case MI_HAVE_DSN:
	for (i = 0; i < I->p.p1->ESMTP_keyword_len; i++)
	    if (0 == istrcmp("DSN",I->p.p1->ESMTP_keywords[i]))
		return 1;
	DPRINT(Debug,10,(&Debug,"DSN not found from EHLO response\n"));
	return 0;
    case MI_DONT_ADD_FROM:
	return (M -> mailer_bits & MB_DONT_ADD_FROM) ? 1 : 0;
    case MI_USE_DOMAIN:
	return (M -> mailer_bits & MB_USE_DOMAIN) ? 1 : 0;
    case MI_HAVE_SENDING_DISPLAY:
	return (M -> mailer_bits & MB_SENDING_DISPLAY) ? 1 : 0;
    }

    return 0; /* Not available */
}

S_(mailer_info_verify_addr smtp_tail_verify_addr)
static int smtp_tail_verify_addr P_((struct mailer_config *M,
				     struct mailer_info  *I, 
				     const char *text,
				     struct addr_verify_result *result));
static int smtp_tail_verify_addr(M,I,text,result)
     struct mailer_config *M;
     struct mailer_info  *I;
     const char *text;
     struct addr_verify_result *result;
{
    int i;
    int skip_enchanced = 0;
   
    if (SMTP_mailer_info_magic != I->p.p1->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_tail_verify_addr",
	      "Bad magic number (smtp mailer info)",0);

    if (-1 == I->p.p1->ESMTP_keyword_len) {

	if (!smtp_run_ehlo(I->p.p1)) {
	    /* Some extension failed hard */
	    DPRINT(Debug,13,(&Debug,
			     "-- Some extensions failed hard -- do not verify address\n"));
	    return -2;   /* caller does job */
	}
    }

    if (I->p.p1->SMTP_flags & SMTP_noverify) {
	DPRINT(Debug,13,(&Debug, "-- Verify temporary disabled\n"));
	return -2;       /* Cannot VRFY user */		
    }

    for (i = 0; i < I->p.p1->ESMTP_keyword_len; i++)
	if (0 == istrcmp("ENHANCEDSTATUSCODES",I->p.p1->ESMTP_keywords[i]))
	    skip_enchanced++;
    
    if (skip_enchanced) {
	DPRINT(Debug,13,(&Debug,
			 "-- Need skip enchanced status codes from responses\n"));
    }

    if (smtp_start_command(I->p.p1,"VRFY")) {
	int succeed = 0;
	char response_code[4];

	smtp_command_arg(I->p.p1,text);
	smtp_command_ready(I->p.p1,0);

	if (smtp_command_ok(I->p.p1,response_code)) {
	    if (0 == memcmp(response_code,"252",3)) {
		DPRINT(Debug,13,(&Debug,
				 " verify response %.3s -- cannot verify user\n",
				 response_code));
		succeed = -2;       /* Cannot VRFY user (relaying?) */
	    } else if (0 == memcmp(response_code,"250",3)) {
		int idx;
		const char * text = NULL;
		const char * buffer = NULL;

		for (idx = 0; NULL != (text = smtp_command_response(I->p.p1,
								    &idx));) {
		    buffer = text;
		}
		
		DPRINT(Debug,13,(&Debug," %d verify responses, result=%s\n",
				 idx,buffer ? buffer : "<NULL>"));

		/* If there was exactly one answer we can use it */
		if (1 == idx) {
		    struct addr_list * X;
		    char *x;
		    
		    if (skip_enchanced && buffer[0] == '2' &&
			(x = strchr(buffer,' '))) {
			buffer = x+1;
			DPRINT(Debug,13,(&Debug," strimmed result=%s\n",
					 buffer));
		    }

		    /* FIXME: Posibble wrong charset */			
		    X =  parse_header_address(NULL,buffer,
					      0,system_charset,NULL);
		    if (X) {
			int len = addr_list_item_count(X);

			if (1 == len) {
			    int group = -1;
			    const struct address * address =
				addr_list_get_item(X,0,&group);

			    const char          * addr     = 
				address_get_ascii_addr(address);
			    const struct string * fullname = 
				address_get_phrase(address);
			    const struct string * comment  = 
				address_get_comment(address);
			    
			    if (addr) {
				result->addr    = safe_strdup(addr);
				
				result->fullname = 
				    fullname ? dup_string(fullname) : NULL;
				result->comment  =
				    comment ? dup_string(comment) : NULL;
				
				succeed = result->fullname ? 1 : -1;
				DPRINT(Debug,13,(&Debug," filled addr=%s\n",
						 result->addr));
			    } else
				succeed = -1;   /* User OK, but caller should fill data */
			} else
			    succeed = -1;   /* User OK, but caller should fill data */

		    free_addr_list(&X);
		    } else
			succeed = -1;   /* User OK, but caller should fill data */

		} else
		    succeed = -1;   /* User OK, but caller should fill data */
	    }
	} else {
	    /* Failed -- see if response code indicates user unknown */
	    if (0 == memcmp(response_code,"553",3) /* Ambiguous */ ||
		0 == memcmp(response_code,"550",3) /* mailbox unavailable */
		) { 
		succeed = 0;		
	    } else if ('4' == response_code[0])
		succeed = -2;       /* Cannot VRFY user (temporray error) */
	    else {
		succeed = -2;       /* Cannot VRFY user ... */
		DPRINT(Debug,13,(&Debug, "-- Disabling verify...\n"));
		I->p.p1->SMTP_flags |= SMTP_noverify;
	    }
	}

	smtp_command_reset(I->p.p1);	
	DPRINT(Debug,13,(&Debug, 
			 "smtp_tail_verify_addr: %s = %d\n",
			 text,succeed));
	return succeed;
    }
    return -2;   /* caller does job */
}

S_(mailer_info_gen_def_ef smtp_common_mailer_info_gen_def_ef)
static void smtp_common_mailer_info_gen_def_ef P_((struct mailer_config *M,
						   struct mailer_info *I,
						   struct mailer_env_from *X)); 
static void smtp_common_mailer_info_gen_def_ef(M,I,X)
     struct mailer_config *M;
     struct mailer_info *I;
     struct mailer_env_from *X; 
{
    if (X->mail_from)
	free(X->mail_from);
    X->mail_from = default_from();
}

S_(mailer_info_set_ef smtp_common_mailer_info_set_ef)
static  void smtp_common_mailer_info_set_ef(M,I,X,value)
     struct mailer_config *M;
     struct mailer_info *I;
     struct mailer_env_from *X;
     const char *value;
{
    if (0 == strcmp(value,"<>")) {
	/* Null envelope sender is only valid as sender (not as recipient)
	   so do not verify it
	*/
	X->mail_from = strmcpy(X->mail_from,value);
	return;
    }

    if (SMTP_mailer_info_magic != I->p.p1->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"smtp_common_mailer_info_set_ef",
	      "Bad magic number (smtp mailer info)",0);

    if (-1 == I->p.p1->ESMTP_keyword_len) {

	if (!smtp_run_ehlo(I->p.p1)) {
	    DPRINT(Debug,4,(&Debug, "-- some extension failed hard -- ignoring now\n"));
	}
    }

    if (I->p.p1->SMTP_flags & SMTP_noverify) {
	DPRINT(Debug,4,(&Debug, "-- Verify temporary disabled - smtp_common_mailer_info_set_ef assuming ok \n"));
	goto succeed;       /* Cannot VRFY user */		
    }

    if (smtp_start_command(I->p.p1,"VRFY")) {
	int succeed;
	char response_code[4];

	succeed  = 0;
	smtp_command_arg(I->p.p1,value);
	smtp_command_ready(I->p.p1,0);

	if (smtp_command_ok(I->p.p1,response_code)) {
	    if (0 == memcmp(response_code,"252",3)) {
		DPRINT(Debug,4,(&Debug,
				 " verify response %.3s -- cannot verify user -- smtp_common_mailer_info_set_ef assuming ok \n",
				 response_code));
		succeed = 1; /* Cannot VRFY user (relaying?) */
	    } else if (0 == memcmp(response_code,"250",3)) {
		succeed = 1;
	    } else {
		DPRINT(Debug,4,(&Debug,
				 " verify response %.3s -- smtp_common_mailer_info_set_ef assuming ok \n",
				 response_code));
		succeed = 1; /* unexpect OK response */
	    }
	} else {
	    /* Failed -- see if response code indicates user unknown */
	    if (0 == memcmp(response_code,"553",3) /* Ambiguous */ ||
		0 == memcmp(response_code,"550",3) /* mailbox unavailable */
		) { 
		DPRINT(Debug,4,(&Debug,
				" verify response %.3s -- cannot verify user -- smtp_common_mailer_info_set_ef failed\n",
				response_code));
		    goto failed;
	    } else if ('4' == response_code[0])
		DPRINT(Debug,4,(&Debug,
				" verify response %.3s -- cannot verify user -- smtp_common_mailer_info_set_ef assuming ok \n",
				response_code));
	    else {		
		DPRINT(Debug,13,(&Debug, "-- Disabling verify...\n"));
		I->p.p1->SMTP_flags |= SMTP_noverify;
		DPRINT(Debug,4,(&Debug,
				" verify response %.3s -- cannot verify user -- smtp_common_mailer_info_set_ef assuming ok \n",
				response_code));
	    }
	    succeed = 1;
	}
        
	if (succeed) {
	succeed:
	    X->mail_from = strmcpy(X->mail_from,value);
	} else {
	failed:
	    lib_error(CATGETS(elm_msg_cat, SmtpSet,
			      SmtpEnvSenderFailed,
			      "Changing of envelope sender failed: %s"),
		      value);
	}
	smtp_command_reset(I->p.p1);	
    } else {
	DPRINT(Debug,4,(&Debug,
			"smtp_common_mailer_info_set_ef failed to run verify command\n"));
	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpEnvSenderFailed,
			  "Changing of envelope sender failed: %s"),
		  value);
    }
}



/* submission mailer ------------------------------------------------------ */

#include <sys/socket.h>

static int submission_verify_tls_certificate = 0;

static char *submission_server = "localhost";

static char REQUIRE_SERVER_NAME;

static char * submission_peer_name = NULL;

static int  submission_verify_peer  P_((struct mailer_info_1 *I,
					struct cancel_data *cd));
static int  submission_verify_peer(I,cd) 
     struct mailer_info_1 *I;
     struct cancel_data *cd;   
{
    int ret = 0;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"submission_verify_peer",
	      "Bad magic number",0);

    /* May need run ehlo */
    if (-1 == I->ESMTP_keyword_len) {
	if (!smtp_run_ehlo_c(I,cd)) {
	    ret = 0; /* Some extension failed hard 
			or wait canceled
		      */
	    goto failed;
	}
    }

    if (submission_verify_tls_certificate) {
	int r = 0;
	StreamInfo(I->RA.stream,SS_verify,&r,NULL,NULL);

	if (r < 0) {
	    lib_error(CATGETS(elm_msg_cat, SmtpSet,
			      SmtpFailVerifyCert,
			      "Failed to verify %S server %s certificate (required)."),
		      smtp_Server(I),
		      I->RA.host);
	    ret = 0;
	    goto failed;
	} else if (0 == r) {
	    int bits;

	    StreamInfo(I->RA.stream,SS_ssf,&bits,NULL,NULL);

	    if (bits > 0)
		lib_error(CATGETS(elm_msg_cat, SmtpSet,
				  SmtpCertNotVerf,
				  "%S server %s certificate not verified (required)."),
			  smtp_Server(I),
			  I->RA.host);
	    else
		lib_error(CATGETS(elm_msg_cat, SmtpSet,
				  SmtpConnectionNotEnc,
				  "%S server %s connection not encrypted (certificate required)."),
			  smtp_Server(I),
			  I->RA.host);
	    ret = 0;
	    goto failed;			      
	}
    }


    if (submission_peer_name) {
	struct string * require_name = NULL;
	struct string * returned_name = NULL;
	int r;

	if (&REQUIRE_SERVER_NAME == submission_peer_name)
	    require_name = new_string2(system_charset,s2us(submission_server));
	else
	    require_name = new_string2(system_charset,
				       s2us(submission_peer_name));

	if (! (r = StreamVerifyName(I->RA.stream,SS_peer_cn,require_name,
				    & returned_name))) {

	    if (returned_name) {
		lib_error(CATGETS(elm_msg_cat, SmtpSet,
				  SmtpCertCnBad,
				  "%S server %s certificate CN is %S, but %S required."),
			  smtp_Server(I),
			  I->RA.host,
			  returned_name,require_name);
	    } else {
		int bits;
		
		StreamInfo(I->RA.stream,SS_ssf,&bits,NULL,NULL);
		
		if (bits > 0)
		    lib_error(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpCertCnReq,
				      "%S server %s certificate CN %S required."),
			      smtp_Server(I),
			      I->RA.host,
			      require_name);
		else
		    lib_error(CATGETS(elm_msg_cat, SmtpSet,
				      SmtpCertNotEncCnReq,
				      "%S server %s connection not encrypted, but certificate CN %S required."),
			      smtp_Server(I),
			      I->RA.host,
			      require_name);
	    }
	}
		
	if (returned_name)			       
	    free_string(&returned_name);
	free_string(&require_name);

	if (!r) {
	    ret = 0;
	    goto failed;
	}
    }

    ret = 1;

 failed:
    DPRINT(Debug,13,(&Debug,"submission_verify_peer=%d\n",
		     ret));

    return ret;
}

/* return 1 if cb success, 0 failure
   return -1 no connection
 */
S_(ra_verify_peer_cb_f submission_verify_peer_cb)
static int submission_verify_peer_cb P_((struct remote_account *ra,
					 union ra_verify_peer_cb_data *data,
					 const struct string * Server,
					 const struct string * server));
static int submission_verify_peer_cb(ra,data,Server,server)
   struct remote_account *ra;
   union ra_verify_peer_cb_data *data;
   const struct string * Server;
   const struct string * server;
{
    int ret = 1;
    struct mailer_info_1 *I = data->mailer;

    /* May need run ehlo */
    if (-1 == I->ESMTP_keyword_len) {
	if (!smtp_run_ehlo(I)) {
	    ret = 0; /* Some extension failed hard */
	}
    }

    DPRINT(Debug,13,(&Debug,"submission_verify_peer_cb=%d\n",
		     ret));

    return ret;
}

static int  submission_verify_peer2  P_((struct mailer_info_1 *I));
static int  submission_verify_peer2(I) 
     struct mailer_info_1 *I;
{
    int status = 1;
    union ra_verify_peer_cb_data cbdata;

    if (SMTP_mailer_info_magic != I->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"submission_verify_peer",
	      "Bad magic number",0);
    cbdata.mailer = I;
    
    if (remote_account_verify_peer(& (I->RA),&cbdata,
				   submission_verify_peer_cb,
				   Server_s((I->SMTP_flags & SMTP_submission) ?
					    Submission_s : SMTP_s),
				   Server_s((I->SMTP_flags & SMTP_submission) ?
					    submission_s : SMTP_s)) < 1) {
	status = 0;   
    }

    DPRINT(Debug,13,(&Debug,"submission_verify_peer2=%d\n",
		     status));

    return status;
}

static int mailer_wait_greeting P_((struct mailer_config *M,
				    struct mailer_info *I,
				    struct cancel_data *cd));
static int mailer_wait_greeting(M,I,cd)
     struct mailer_config *M;
     struct mailer_info *I;
     struct cancel_data *cd;
{
    struct mailer_info_1 *I1 = I->p.p1;


    if (SMTP_mailer_info_magic != I1->magic)
	    panic("MAILER PANIC",__FILE__,__LINE__," mailer_wait_greeting",
		  "Bad magic (number mailer_info_1)",0);

    I1->state = smtp_wait_reply;

    ConfigStream(I1->RA.stream,
		 smtp_read_action,ss_noaction_routine,smtp_timeout_action,
		 SMTP_TIMEOUT,I1);
  
    while ( smtp_wait_reply  == I1->state &&
	    NULL         != I1->RA.stream) {

	if (!WaitStreamFor_c(I1->RA.stream,SS_read_act,cd)) {
	    
	    DPRINT(Debug,7,(&Debug,"mailer_wait_greeting: wait not succeed?\n"));

	    if (cd && is_canceled(cd)) {
		DPRINT(Debug,7,(&Debug,"mailer_wait_greeting: wait canceled!\n"));

		return 0;
	    }
	}
    }

    if (smtp_reply_incomplete == I1->state ||
	smtp_reply_read     == I1->state) {
	
	DPRINT(Debug,7,(&Debug,"mailer_wait_greeting: response code %.3s\n",
			I1->response_code));

	if ('2' == I1->response_code[0]) {
	    DPRINT(Debug,7,(&Debug,"mailer_wait_greeting: OK\n"));
	    return 1;
	}
    }

    DPRINT(Debug,7,(&Debug,"mailer_wait_greeting: failed\n"));
    return 0;
}

S_(mailer_info_init_hook submission_mailer_info_init)
static int submission_mailer_info_init P_((struct mailer_config *M,
					   struct mailer_info *I));
static int submission_mailer_info_init(M,I)
     struct mailer_config *M;
     struct mailer_info *I;
{
    struct service_entry *se;
    int stflag;

    PORTS ports[] = { PORT_submission, PORT_smtp, PORT_end };
    enum itls_status have_initial_tls = itls_none;
    int got;  /* port number */
    int status = 0;   
    struct cancel_data * cd = NULL;

    DPRINT(Debug,13,(&Debug,
		     "submission_mailer_info_init: mailer info, I=%p\n",I));

    se = give_service_entry(submission_server, STFLAG_smtp_like);
    if (!se)
	return 0;

    I->p.p1 = safe_malloc(sizeof (* (I->p.p1)));

    /* bzero is defined on hdrs/defs.h */
    bzero((void *)I->p.p1,sizeof (* (I->p.p1)));

    I->p.p1->magic = SMTP_mailer_info_magic;

    zero_remote_account(& (I->p.p1->RA));

    /* There is no connection caching currently */

    I->p.p1->RA.host = strmcpy(I->p.p1->RA.host,se->official_name);
    		 
    zero_Read_Buffer(& (I->p.p1->read_buffer));
    zero_Write_Buffer(& (I->p.p1->write_buffer));
    I->p.p1->response_code[0]  = '\0';
    I->p.p1->response_text     = NULL;
    I->p.p1->response_text_len = 0;

    I->p.p1->incomplete_command     = NULL;
    I->p.p1->incomplete_command_len = 0;

    I->p.p1->ESMTP_keywords        = NULL;
    I->p.p1->ESMTP_keyword_len     = -1;   /* -1 == EHLO not issued yet */
    I->p.p1->state                 = smtp_error;

    I->p.p1->SMTP_flags        = 0;


    if (!connect_remote_account(&(I->p.p1->RA),&got,se,ports,
				PORT_end))
	goto clean;

    switch((have_initial_tls = remote_account_init_tls(&(I->p.p1->RA)))) {
	
    case itls_unsupported:
    case itls_failed:
	status =  0;
	goto clean;
	
    case itls_none:
	break;
	
    case itls_done:
	DPRINT(Debug,8,(&Debug,
			"submission_mailer_info_init: Initial TLS done (no STARTTLS)\n"));
    }
 

    switch((stflag = STFLAG_service_type(se))) {
    case STFLAG_is_submission:
	goto is_submission;
    case STFLAG_is_smtp:
	goto is_smtp;
    default:
	DPRINT(Debug,8,(&Debug,
			"submission_mailer_info_init: unknown stflag=x%04x (service=%x, port=%d)\n",
			stflag,se->service,got));
    }

    switch (got) {
    case PORT_submission:
    is_submission:
    
	DPRINT(Debug,8,(&Debug,
			"submission_mailer_info_init: Submission service\n"));
	I->p.p1->SMTP_flags    = SMTP_submission;
	break;

    case PORT_smtp:
    is_smtp:

	DPRINT(Debug,8,(&Debug,
			"submission_mailer_info_init: smtp service\n"));
	    
	break;

    default:
	DPRINT(Debug,1,(&Debug,
			"submission_mailer_info_init: Unknown service!\n"));
    }

    if (I->p.p1->RA.hostname && 0 != strcmp(se->official_name,I->p.p1->RA.hostname))
	cd = new_schedule_cancel(cancel_delay_msec,
				 CATGETS(elm_msg_cat, SmtpSet,
					 SmtpWaitingServer1,
					 "Waiting %S server %s [%s], port %d ..."),
				 Server_s((I->p.p1->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s),
				 se->official_name,
				 I->p.p1->RA.hostname,
				 got);			
    else
	cd = new_schedule_cancel(cancel_delay_msec,
				 CATGETS(elm_msg_cat, SmtpSet,
					 SmtpWaitingServer,
					 "Waiting %S server %s, port %d ..."),
				 Server_s((I->p.p1->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s),
				 se->official_name,
				 got);
    
    if (!mailer_wait_greeting(M,I,cd)) {
	
	DPRINT(Debug,12,(&Debug,
			 "submission_mailer_info_init: Failed to get greeting from server\n"));
	
	/* Caller will call submission_mailer_info_close ... */
	status =  0;
	goto clean;
    }
    if (!smtp_command_reset_c(I->p.p1,cd)) {

	DPRINT(Debug,12,(&Debug,
			 "submission_mailer_info_init: command reset failed\n"));

	status =  0;
	goto clean;
    }
   
    status = 1;

    if (submission_verify_tls_certificate ||
	submission_peer_name)
	status = submission_verify_peer(I->p.p1,cd);
    else {
	DPRINT(Debug,12,(&Debug,
			 "submission_mailer_info_init: No verify-tls-certificate or require-tls-peer-name (on mailer defination)\n"));
    }
    free_cancel(&cd);

    if (status) 
	status = submission_verify_peer2(I->p.p1);

 clean:
    if (cd)
	free_cancel(&cd);

    free_service_entry(&se);

    DPRINT(Debug,12,(&Debug,
		     "submission_mailer_info_init=%d\n",status));
    return status;
}

static void mailer_close2 P_((struct mailer_info *I));
static void mailer_close2(I)
     struct mailer_info *I;
{
    if (SMTP_mailer_info_magic != I->p.p1->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"mailer_close2",
	      "Bad magic number (smtp mailer info)",0);

    free_Read_Buffer(& (I->p.p1->read_buffer));
    free_Write_Buffer(& (I->p.p1->write_buffer));
    
    free_response_text(I->p.p1);
    
    if (I->p.p1->incomplete_command)
	free(I->p.p1->incomplete_command);
    
    I->p.p1->incomplete_command     = NULL;
    I->p.p1->incomplete_command_len = 0;
    
    free_ESMTP_keywords(I->p.p1);
}

/* Return
   1 == mailer restarted, reconnect succeed -- reask info
   0 == mailer not disconnected
  -1 == mailer disconnected, reconnect failed 
*/
S_(mailer_info_restart_hook submission_mailer_info_rs_hook)
static int submission_mailer_info_rs_hook P_((struct mailer_config *M,
					 struct mailer_info *I));
static int submission_mailer_info_rs_hook(M,I)
     struct mailer_config *M;
     struct mailer_info *I;
{
    int is_ok = 1;
    int status = 0;
    int stflag;
    struct cancel_data * cd = NULL;

    PORTS ports[] = { PORT_submission, PORT_smtp, PORT_end };
    int got;  /* port number */
    enum itls_status have_initial_tls = itls_none;
    struct service_entry *se = NULL;

    if (SMTP_mailer_info_magic != I->p.p1->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,
	      "submission_mailer_info_rs_hook",
	      "Bad magic number (smtp mailer info)",0);

    if (I->p.p1->RA.hostname && 0 != strcmp(I->p.p1->RA.host,I->p.p1->RA.hostname)) 
	cd = new_schedule_cancel(cancel_delay_msec,
				 CATGETS(elm_msg_cat, SmtpSet,
					 SmtpWaitingServer3,
					 "Waiting %S server %s [%s] ..."),
				 Server_s((I->p.p1->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s),
				 I->p.p1->RA.host,
				 I->p.p1->RA.hostname);
    
    else
	cd = new_schedule_cancel(cancel_delay_msec,
				 CATGETS(elm_msg_cat, SmtpSet,
					 SmtpWaitingServer2,
					 "Waiting %S server %s ..."),
				 Server_s((I->p.p1->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s),
				 I->p.p1->RA.host);
    
    if (smtp_start_command_c(I->p.p1,"NOOP",cd)) {
	char response_code[4];

	if (!smtp_command_ready_c(I->p.p1,0,cd)) {
	    DPRINT(Debug,12,(&Debug,
			     "submission_mailer_info_rs_hook: NOOP not ready\n"));
	    goto noop_failure;
	}

	is_ok = smtp_command_ok_c(I->p.p1,response_code,cd);

	if (!smtp_command_reset_c(I->p.p1,cd)) {
	    DPRINT(Debug,12,(&Debug,
			     "submission_mailer_info_rs_hook: NOOP -- command reset failure\n"));
	    is_ok = 0;
	}
    } else {
    noop_failure:
	is_ok = 0;
    }

    if (is_ok) {
	DPRINT(Debug,12,(&Debug,
			 "submission_mailer_info_rs_hook=0  (OK, mailer not disconnected)\n"));
	
	free_cancel(&cd);

	return 0;   /* Mailer OK -- need not restart */
    }

    /* Send QUIT */
    if (smtp_idle == I->p.p1->state) {
	int status = smtp_start_command_c(I->p.p1,"QUIT",cd);
	
	if (status) {
	    if (!smtp_command_ready_c(I->p.p1,0,cd)) {
		DPRINT(Debug,12,(&Debug,
				 "submission_mailer_info_rs_hook: QUIT not ready\n"));
		goto quit_failure;
	    }
	    
	    /* Wait for QUIT */	       
	    if (!smtp_command_reset_c(I->p.p1,cd)) {
		DPRINT(Debug,12,(&Debug,
				 "submission_mailer_info_rs_hook: QUIT -- command reset failure\n"));
		goto quit_failure;
	    }
	}
    }

 quit_failure:
    status = -1;

    /* Actually close connection */
    clear_remote_account(& (I->p.p1->RA));	
    mailer_close2(I);
    free_cancel(&cd);

    se = give_service_entry(submission_server, STFLAG_smtp_like);
    if (!se) {
        status = -1;
	goto clean;
    }

    /* clear_remote_account was cleared this */
    I->p.p1->RA.host = strmcpy(I->p.p1->RA.host,se->official_name);

    /* Try reconnect */
    I->p.p1->state                 = smtp_error;
    I->p.p1->SMTP_flags            = 0;

    if (!connect_remote_account(&(I->p.p1->RA),&got,se,ports,
				PORT_end))
	goto clean;
    
    switch((have_initial_tls = remote_account_init_tls(&(I->p.p1->RA)))) {
	
	case itls_unsupported:
	case itls_failed:
	    
	    goto clean;

	case itls_none:
	    break;

	case itls_done:
	    DPRINT(Debug,8,(&Debug,
			    "submission_mailer_info_rs_hook: Initial TLS done (no STARTTLS)\n"));
	}



    switch((stflag = STFLAG_service_type(se))) {
    case STFLAG_is_submission:
	    goto is_submission;
    case STFLAG_is_smtp:
	goto is_smtp;

    default:
	DPRINT(Debug,8,(&Debug,
			"submission_mailer_info_rs_hook: unknown stflag=x%04x (service=%x, port=%d)\n",
			stflag,se->service,got));
    }
    
    switch (got) {
    case PORT_submission:
    is_submission:
	
	DPRINT(Debug,8,(&Debug,
			"submission_mailer_info_rs_hook: Submission service\n"));
	I->p.p1->SMTP_flags    = SMTP_submission;
	break;

    case PORT_smtp:
    is_smtp:

	DPRINT(Debug,8,(&Debug,
			"submission_mailer_info_rs_hook: smtp service\n"));
	    
	break;
    default:
	DPRINT(Debug,1,(&Debug,
			"submission_mailer_info_rs_hook: Unknown service!\n"));
    }

    if (cd)
	free_cancel(&cd);

    if (I->p.p1->RA.hostname && 0 != strcmp(se->official_name,I->p.p1->RA.hostname))
	cd = new_schedule_cancel(cancel_delay_msec,
				 CATGETS(elm_msg_cat, SmtpSet,
					 SmtpWaitingServer1,
					 "Waiting %S server %s [%s], port %d ..."),
				 Server_s((I->p.p1->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s),
				 se->official_name,
				 I->p.p1->RA.hostname,
				 got);			
    else
	cd = new_schedule_cancel(cancel_delay_msec,
				 CATGETS(elm_msg_cat, SmtpSet,
					 SmtpWaitingServer,
					 "Waiting %S server %s, port %d ..."),
				 Server_s((I->p.p1->SMTP_flags & SMTP_submission) ? submission_s : SMTP_s),
				 se->official_name,
				 got);

    if (!mailer_wait_greeting(M,I,cd)) {
	DPRINT(Debug,12,(&Debug,
			 "submission_mailer_info_rs_hook: Failed to get greeting from server\n"));

	/* Caller will call submission_mailer_info_close ... */
	status =  -1;
	goto clean;
    }
    smtp_command_reset(I->p.p1);
    status = 1;   /* reconnect succeed */

    if (submission_verify_tls_certificate ||
	submission_peer_name) {
	
	if (! submission_verify_peer(I->p.p1,cd)) {

	    if (cd && is_canceled(cd)) {
		DPRINT(Debug,7,(&Debug,"submission_mailer_info_rs_hook: wait canceled!\n"));
		
		status = 0;
		goto clean;
	    }

	    status =  -1;

	}
    } else {
	DPRINT(Debug,12,(&Debug,
			 "submission_mailer_info_rs_hook: No verify-tls-certificate or require-tls-peer-name (on mailer defination)\n"));
    }
    free_cancel(&cd);

    if (status > 0) { 
	if (!submission_verify_peer2(I->p.p1))
	    status =  -1;
    }

 clean:
    if (cd)
	free_cancel(&cd);

    if (se)
	free_service_entry(&se);

    DPRINT(Debug,12,(&Debug,
		     "submission_mailer_info_rs_hook=%d\n",status));
    return status;
}


S_(mailer_info_close_hook submission_mailer_info_close)
static void submission_mailer_info_close P_((struct mailer_config *M,
					     struct mailer_info *I));
static void submission_mailer_info_close(M,I)
     struct mailer_config *M;
     struct mailer_info *I;
{

    if (SMTP_mailer_info_magic != I->p.p1->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"submission_mailer_info_close",
	      "Bad magic number (smtp mailer info)",0);

    DPRINT(Debug,13,(&Debug,
		     "submission_mailer_info_close: mailer info, I=%p\n",I));

    if (I->p.p1) {
	if (smtp_idle == I->p.p1->state) {
	    int status = smtp_start_command(I->p.p1,"QUIT");

	    if (status) {
		smtp_command_ready(I->p.p1,0);
		
		/* Wait for QUIT */	       
		smtp_command_reset(I->p.p1);
	    }
	}

	clear_remote_account(& (I->p.p1->RA));	
	mailer_close2(I);

	I->p.p1->magic = 0; /* Invalidate */

	free(I->p.p1);
	I->p.p1 = NULL;
    } 
}

static int submission_verify_address = 0;


S_(mailer_info_verify_addr submission_mailer_info_verify_addr)
static int submission_mailer_info_verify_addr P_((struct mailer_config *M,
						  struct mailer_info  *I, 
						  const char *text,
						  struct addr_verify_result 
						  *result));
static int submission_mailer_info_verify_addr(M,I,text,result)
     struct mailer_config *M;
     struct mailer_info  *I;
     const char *text;
     struct addr_verify_result *result;
{
    if (!submission_verify_address) {
	DPRINT(Debug,13,(&Debug, 
			 "submission_mailer_info_verify_addr=-2 (Verify disabled)\n"));
	return -2;       /* Cannot VRFY user */		
    }

    return smtp_tail_verify_addr(M,I,text,result);
}

S_(MO_value_hook MO_submission_verify)
static int MO_submission_verify P_((struct mailer_config      * M,
			       struct mailer_option_list * L,
			       char **value,int set));
static int MO_submission_verify (M,L,value,set)
     struct mailer_config      *M;
     struct mailer_option_list *L;
     char **value;
     int set;
{
    static char *no_yes[] = { "no", "yes" };
    static int value_set  = 0;

    if (set) {
	if (!*value)
	     submission_verify_address = 1;
	else if (0 == strcmp(*value,no_yes[0]))
	    submission_verify_address = 0;
	else if (0 == strcmp(*value,no_yes[1])) 
	    submission_verify_address = 1;
	else {
	    lib_error(CATGETS(elm_msg_cat, SmtpSet,
			      SmtpBadConfSubmission,
			      "Bad %s=%s for submission"),
		      L->field,*value);
	    return 0;
	}	
	value_set = 1;
    } else {
	if (submission_verify_address) {
	    *value = no_yes[1];
	    return 1;
	}
	*value = no_yes[0]; /* Only show if verify=no or
			       verify=yes is given at least once 
			    */
    }    
    return value_set;
}

S_(MO_value_hook MO_submission_tlsverify)
static int MO_submission_tlsverify P_((struct mailer_config      * M,
			       struct mailer_option_list * L,
			       char **value,int set));
static int MO_submission_tlsverify (M,L,value,set)
     struct mailer_config      *M;
     struct mailer_option_list *L;
     char **value;
     int set;
{
    static char *no_yes[] = { "no", "yes" };
    static int value_set  = 0;

    if (set) {
	if (!*value)
	     submission_verify_tls_certificate = 1;
	else if (0 == strcmp(*value,no_yes[0]))
	    submission_verify_tls_certificate = 0;
	else if (0 == strcmp(*value,no_yes[1])) 
	    submission_verify_tls_certificate = 1;
	else {
	    lib_error(CATGETS(elm_msg_cat, SmtpSet,
			      SmtpBadConfSubmission,
			      "Bad %s=%s for submission"),
		      L->field,*value);
	    return 0;
	}	
	value_set = 1;
    } else {
	if (submission_verify_tls_certificate) {
	    *value = no_yes[1];
	    return 1;
	}
	*value = no_yes[0]; /* Only show if verify-tls-certificate=no or
			       verify-tls-certificate=yes is given at least once 
			    */
    }    
    return value_set;
}
S_(MO_value_hook MO_submission_requirename)
static int MO_submission_requirename P_((struct mailer_config      * M,
			       struct mailer_option_list * L,
			       char **value,int set));
static int MO_submission_requirename (M,L,value,set)
     struct mailer_config      *M;
     struct mailer_option_list *L;
     char **value;
     int set;
{
    static int value_set  = 0;
   
     if (set) {

	 if (L->malloced) {
	     free(*(L->value));
	     *(L->value) = NULL;
	 }

	 if (!value || !*value) {
	     *(L->value) = &REQUIRE_SERVER_NAME;
	     L->malloced = 0;
	 } else {
	     *(L->value) = safe_strdup(*value);
	     L->malloced = 1;
	 }
	 value_set = 1;
     } else {
	 if (!*(L->value) || *(L->value) == &REQUIRE_SERVER_NAME)
	     *value = NULL;
	 else {
	     *value = *(L->value);
	     return 1;
	 }
     }

     return value_set;
}

S_(MO_value_hook MO_smtp_sending_display)
static int MO_smtp_sending_display P_((struct mailer_config      * M,
				  struct mailer_option_list * L,
				  char **value,int set));
static int MO_smtp_sending_display (M,L,value,set)
     struct mailer_config      *M;
     struct mailer_option_list *L;
     char **value;
     int set;
{
    static char *no_yes[] = { "no", "yes" };
    static int value_set  = 0;

    if (set) {
	if (!*value)
	    M -> mailer_bits  |= MB_SENDING_DISPLAY;
	else if (0 == strcmp(*value,no_yes[0]))
	    M -> mailer_bits  &= ~MB_SENDING_DISPLAY;
	else if (0 == strcmp(*value,no_yes[1])) 
	    M -> mailer_bits  |= MB_SENDING_DISPLAY;
	else {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, 
			      ElmBadSendingDisplay,
			      "Bad sending-display=%s for %s"),
		      *value, M->mailer_type);
	    return 0;
	}

	value_set = 1;
    } else {
	
	*value = no_yes[ (M -> mailer_bits & MB_SENDING_DISPLAY) ?
			 1 : 0 ];
       
	if (0 == (M -> mailer_bits & MB_SENDING_DISPLAY))
	    return 1;

	/* Do not show sending-display=yes if it is not set explicity */
    }

    return value_set;
}


static struct mailer_option_list SUBMISSION_options[] = {
    { "server",   &submission_server,                MO_default_value, 0 },
    { "verify",   NULL,                              MO_submission_verify, 0 },
    { "allow-set-sender",NULL,                       MO_allow_set_sender, 0 },
    { "verify-tls-certificate", NULL,                MO_submission_tlsverify, 
      0 },
    { "require-tls-peer-name", & submission_peer_name, 
      MO_submission_requirename, 0 },
    { "sending-display", NULL,                       MO_smtp_sending_display, 
      0 },
    { NULL, NULL,                                    MO_default_value, 0 }
};

/* sendmail-bs mailer ----------------------------------------------------- */

#if defined(BACKGROUD_PROCESSES) 
#define HAVE_SENDMAILBS 1
#endif

#if HAVE_SENDMAILBS

static char * sendmail_mailer_path    = SENDMAIL_MAILER_PATH;



S_(mailer_info_init_hook sendmailbs_mailer_info_init)
static int sendmailbs_mailer_info_init P_((struct mailer_config *M,
					   struct mailer_info *I));
static int sendmailbs_mailer_info_init(M,I)
     struct mailer_config *M;
     struct mailer_info *I;
{
    const char * argv[3];
    char response_code[4];

    int options = SY_ENV_SHELL | SY_NOTTY;
    int ret,r;
    int pair[2];

    I->p.p1 = safe_malloc(sizeof (* (I->p.p1)));

    /* bzero is defined on hdrs/defs.h */
    bzero((void *)I->p.p1,sizeof (* (I->p.p1)));

    I->p.p1->magic = SMTP_mailer_info_magic;

    DPRINT(Debug,13,(&Debug,
		     "sendmailbs_mailer_info_init: mailer info, I=%p\n",I));

    zero_remote_account(& (I->p.p1->RA));	
    
    r = socketpair(AF_UNIX,SOCK_STREAM,0, pair);
    
    if (-1 == r) {
	int err = errno;
	lib_error(CATGETS(elm_msg_cat, SmtpSet,
			  SmtpSocketPairFail,
			  "socketpair failed: %s"),
		  strerror(err));

	I->p.p1->magic = 0; /* Invalidate */

	free(I->p.p1);
	I->p.p1 = NULL;
	return 0;
    }

    argv[0] = sendmail_mailer_path;
    argv[1] = "-bs";
    argv[2] = NULL;
	
    /* Connect standard imput and output to other end of socket ... */

    ret = start_run( & (I->p.p1->RS),options,argv,pair[1],pair[1]);

    /* Close other end of socket so it is used only by sendmail */
    close(pair[1]);
    pair[1] = -1;

    if (!ret) {
	if (I->p.p1->RS.save_errno)
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailErrno,
			      "Failed: %.30s: %.40s"),
		      argv[0],strerror(I->p.p1->RS.save_errno));
	else
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmCantStart,
			      "Can't start %.30s"),
		      argv[0]);

	close(pair[0]); 
	pair[0] = -1;


	I->p.p1->magic = 0; /* Invalidate */
	free(I->p.p1);
	I->p.p1 = NULL;
	return 0;
    }

    if (transaction_file) {
	time_t tmp = time(NULL);
	struct tm * zz = localtime(&tmp);
	fprintf(transaction_file,
		"%d [%d] %02d:%02d:%02d === CONNECT command %s, pid %d\n",
		getpid(),pair[0] ,
		zz ? zz->tm_hour : 00,
		zz ? zz->tm_min  : 00,
		zz ? zz->tm_sec  : 00,
		sendmail_mailer_path,
		I->p.p1->RS.pid);
    }

    if (-1 == fcntl(pair[0],F_SETFL,O_NONBLOCK)) {
	int err UNUSED_VAROK = errno;

	DPRINT(Debug,12,(&Debug,
			 "sendmailbs_mailer_info_init: fcntl [O_NONBLOCK] failed: %s\n",
			 strerror(err)));
    }

    I->p.p1->RA.stream = returnSimpleStream(pair[0]);
    I->p.p1->state = smtp_wait_reply;

    ConfigStream(I->p.p1->RA.stream,
		 smtp_read_action,ss_noaction_routine,smtp_timeout_action,
		 SMTP_TIMEOUT,I->p.p1);
		 
    zero_Read_Buffer(& (I->p.p1->read_buffer));
    zero_Write_Buffer(& (I->p.p1->write_buffer));
    I->p.p1->response_code[0]  = '\0';
    I->p.p1->response_text     = NULL;
    I->p.p1->response_text_len = 0;

    I->p.p1->incomplete_command     = NULL;
    I->p.p1->incomplete_command_len = 0;

    I->p.p1->ESMTP_keywords        = NULL;
    I->p.p1->ESMTP_keyword_len     = -1;   /* -1 == EHLO not issued yet */

    I->p.p1->SMTP_flags        = SMTP_submission;

    if (!smtp_command_ok(I->p.p1,response_code)) {
	DPRINT(Debug,12,(&Debug,
			 "sendmailbs_mailer_info_init: Failed to get greeting from sendmail -bs\n"));

	/* Caller will call sendmailbs_mailer_info_close ... */
	return 0;
    }
    smtp_command_reset(I->p.p1);

    return 1;
}


/* Return
   1 == mailer restarted, reconnect succeed -- reask info
   0 == mailer not disconnected
  -1 == mailer disconnected, reconnect failed 
*/
S_(mailer_info_restart_hook sendmailbs_mailer_info_rs_hook)
static int sendmailbs_mailer_info_rs_hook P_((struct mailer_config *M,
					 struct mailer_info *I));
static int sendmailbs_mailer_info_rs_hook(M,I)
     struct mailer_config *M;
     struct mailer_info *I;
{
    int is_ok = 0;

    if (SMTP_mailer_info_magic != I->p.p1->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"sendmailbs_mailer_info_rs_hook",
	      "Bad magic number (smtp mailer info)",0);
    
    if (smtp_start_command(I->p.p1,"NOOP")) {
	char response_code[4];

	smtp_command_ready(I->p.p1,0);

	is_ok = smtp_command_ok(I->p.p1,response_code);

	smtp_command_reset(I->p.p1);
    } else
	is_ok = 0;

    if (is_ok) {
	DPRINT(Debug,12,(&Debug,
			 "sendmailbs_mailer_info_rs_hook=0  (OK, mailer not disconnected)\n"));
	
	return 0;   /* Mailer OK -- need not restart */
    }

    /* TODO -- implement reconnect */

    DPRINT(Debug,4,(&Debug,
		     "sendmailbs_mailer_info_rs_hook=-1 (reconnect not implemented)\n"));

    return -1;  /* reconnect not implemented */
}


S_(mailer_info_close_hook sendmailbs_mailer_info_close)
static void sendmailbs_mailer_info_close(M,I)
     struct mailer_config *M;
     struct mailer_info *I;
{

    if (SMTP_mailer_info_magic != I->p.p1->magic)
	panic("MAILER PANIC",__FILE__,__LINE__,"sendmailbs_mailer_info_close",
	      "Bad magic number (smtp mailer info)",0);

    DPRINT(Debug,13,(&Debug,
		     "sendmailbs_mailer_info_close: mailer info, I=%p\n",I));

    if (I->p.p1) {
	int ret,exit_code;

	ret = run_already_done(& (I->p.p1->RS),&exit_code);

	if (0 == ret && smtp_idle == I->p.p1->state) {
	    int status = smtp_start_command(I->p.p1,"QUIT");

	    if (status) {
		smtp_command_ready(I->p.p1,0);
		
		/* Wait for QUIT */	       
		smtp_command_reset(I->p.p1);
	    }
	}

	clear_remote_account(& (I->p.p1->RA));	
	
	if (0 == ret) 
	    ret = wait_end(&(I->p.p1->RS),&exit_code);

	if (ret < 0) {
	    DPRINT(Debug,4,(&Debug,
			    "%s died on signal %d\n",
			    sendmail_mailer_path,-ret));

	    if (transaction_file) {
		time_t tmp = time(NULL);
		struct tm * zz = localtime(&tmp);
		fprintf(transaction_file,
			"%d [?] %02d:%02d:%02d === DIED command %s, pid %d, signal %d\n",
			getpid(),
			zz ? zz->tm_hour : 00,
			zz ? zz->tm_min  : 00,
			zz ? zz->tm_sec  : 00,
			sendmail_mailer_path,
			I->p.p1->RS.pid,
			-ret);
	    }
	} else if (0 == ret) {
	    DPRINT(Debug,4,(&Debug,
			    "%s lost?\n",
			    sendmail_mailer_path));

	    if (transaction_file) {
		time_t tmp = time(NULL);
		struct tm * zz = localtime(&tmp);
		fprintf(transaction_file,
			"%d [?] %02d:%02d:%02d === LOST command %s, pid %d\n",
			getpid(),
			zz ? zz->tm_hour : 00,
			zz ? zz->tm_min  : 00,
			zz ? zz->tm_sec  : 00,
			sendmail_mailer_path,
			I->p.p1->RS.pid);
	    }
	} else {
	    DPRINT(Debug,4,(&Debug,
			    "%s exited with %d%s\n",
			    sendmail_mailer_path,
			    exit_code, exit_code == 0 ? " (OK)" : ""));

	    if (transaction_file) {
		time_t tmp = time(NULL);
		struct tm * zz = localtime(&tmp);
		fprintf(transaction_file,
			"%d [?] %02d:%02d:%02d === EXITED command %s, pid %d, exit code %d%s\n",
			getpid(),
			zz ? zz->tm_hour : 00,
			zz ? zz->tm_min  : 00,
			zz ? zz->tm_sec  : 00,
			sendmail_mailer_path,
			I->p.p1->RS.pid,
			exit_code,
			exit_code == 0 ? " (OK)" : "");
	    }
	}

	mailer_close2(I);

	I->p.p1->magic = 0; /* Invalidate */

	free(I->p.p1);
	I->p.p1 = NULL;
    } 
}

static int sendmailbs_verify_address = 0;

S_(mailer_info_verify_addr sendmailbs_mailer_info_verify_addr)
static int sendmailbs_mailer_info_verify_addr(M,I,text,result)
     struct mailer_config *M;
     struct mailer_info  *I;
     const char *text;
     struct addr_verify_result *result;
{
    if (!sendmailbs_verify_address) {
	DPRINT(Debug,13,(&Debug, 
			 "sendmailbs_mailer_info_verify_addr=-2 (Verify disabled)\n"));
	return -2;       /* Cannot VRFY user */		
    }

    return smtp_tail_verify_addr(M,I,text,result);
}

S_(MO_value_hook MO_sendmailbs_verify)
static int MO_sendmailbs_verify P_((struct mailer_config      * M,
			       struct mailer_option_list * L,
			       char **value,int set));
static int MO_sendmailbs_verify (M,L,value,set)
     struct mailer_config      *M;
     struct mailer_option_list *L;
     char **value;
     int set;
{
    static char *no_yes[] = { "no", "yes" };
    static int value_set  = 0;

    if (set) {
	if (!*value)
	     sendmailbs_verify_address = 1;
	else if (0 == strcmp(*value,no_yes[0]))
	    sendmailbs_verify_address = 0;
	else if (0 == strcmp(*value,no_yes[1])) 
	    sendmailbs_verify_address = 1;
	else {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSendmailBadVerify,
			      "Bad verify=%s for sendmail"),
		      *value);
	    return 0;
	}	
	value_set = 1;
    } else {
	if (sendmailbs_verify_address) {
	    *value = no_yes[1];
	    return 1;
	}
	*value = no_yes[0]; /* Only show if verify=no or
			       verify=yes is given at least once 
			    */
    }    
    return value_set;
}


static struct mailer_option_list SENDMAILBS_options[] = {
    { "path",     &sendmail_mailer_path,             MO_default_value, 0 },
    { "verify",   NULL,                              MO_sendmailbs_verify, 0 },
    { "allow-set-sender",NULL,                       MO_allow_set_sender,  0 },
    { "sending-display", NULL,                       MO_smtp_sending_display, 
      0 },
    { NULL, NULL,                                    MO_default_value, 0 }
};

#endif

static struct mailer_config MAILERS[] = {
    { MAILER_CONFIG_magic,
      "submission", &(SUBMISSION_options[0]), default_set_option, 
      smtp_common_mailer_init, smtp_common_mailer_close, 
      smtp_common_mailer_backend3,
      NULL, 
      submission_mailer_info_init, submission_mailer_info_close,
      smtp_common_mailer_info_query, submission_mailer_info_verify_addr,
      smtp_common_mailer_info_gen_def_ef, smtp_common_mailer_info_set_ef,
      submission_mailer_info_rs_hook,
      
      MB_USE_DOMAIN|MB_REQ_DEFAULT_SENDER|MB_SENDING_DISPLAY
    },
#if HAVE_SENDMAILBS
    { MAILER_CONFIG_magic,
      "sendmail-bs", &(SENDMAILBS_options[0]), default_set_option, 
      smtp_common_mailer_init, smtp_common_mailer_close, 
      smtp_common_mailer_backend3,
      &sendmail_mailer_path, 
      sendmailbs_mailer_info_init, sendmailbs_mailer_info_close,
      smtp_common_mailer_info_query, sendmailbs_mailer_info_verify_addr,
      smtp_common_mailer_info_gen_def_ef, smtp_common_mailer_info_set_ef,
      sendmailbs_mailer_info_rs_hook,

      MB_USE_DOMAIN|MB_REQ_DEFAULT_SENDER|MB_SENDING_DISPLAY
    },
#endif

};

#if ANSI_C
provides_shared_MCF_f provides_shared_MCF;
#endif
struct mailer_config * provides_shared_MCF P_((int *count, size_t *s_size, 
					       size_t *opt_size,
					       size_t *s_send_state_size, 
					       size_t *s_cancel_size,
					       size_t *s_mailer_info_size, 
					       size_t *s_env_from_size));
struct mailer_config * provides_shared_MCF(count,s_size,opt_size,
					   s_send_state_size,s_cancel_size,
					   s_mailer_info_size,s_env_from_size)
     int *count; 
     size_t *s_size; 
     size_t *opt_size;
     size_t *s_send_state_size;
     size_t *s_cancel_size;
     size_t *s_mailer_info_size;
     size_t *s_env_from_size;
{
    *s_size    = sizeof (MAILERS[0]);
    *count  = (sizeof MAILERS) / sizeof (MAILERS[0]);

    *opt_size           = sizeof (struct mailer_option_list); 
    *s_send_state_size  = sizeof (struct mail_send_state);
    *s_cancel_size      = sizeof (struct mailer_cancel_ticket);
    *s_mailer_info_size = sizeof (struct mailer_info);
    *s_mailer_info_size = sizeof (struct mailer_info);
    *s_env_from_size    = sizeof (struct mailer_env_from);

    if (!remote_account_OK(sizeof (struct remote_account))) {

	DPRINT(Debug,1,(&Debug,			    
			"smtp:provides_shared_MCF: sizeof (struct remote_account) mismatch\n"));

	*count = 0;
    }

    if (! StreamOK(sizeof (struct stream_type))) {
	DPRINT(Debug,1,(&Debug,			    
			"smtp:provides_shared_MCF: sizeof (struct stream_type) mismatch\n"));

	*count = 0;
    }

    return &(MAILERS[0]);
}

static struct string * static_strings [ NUM_s ];
static const struct string * Server_s(idx)
     enum server_s idx;
{
    if (idx < 0 || idx >= NUM_s)
	panic("MAILER PANIC",__FILE__,__LINE__,"Server_s",
	      "Bad idx",0);	
    
    if (static_strings[idx])
	return static_strings[idx];

    switch (idx) {
    case Submission_s:
	static_strings[idx] = 
	    format_string(CATGETS(elm_msg_cat, SmtpSet,
				  SmtpSubmission,
				  "Submission"));
	break;
    case SMTP_s:
	static_strings[idx] = 
	    format_string(CATGETS(elm_msg_cat, SmtpSet,
				  SmtpSMTP,
				  "SMTP"));
	break;
    case submission_s:
	static_strings[idx] = 
	    format_string(CATGETS(elm_msg_cat, SmtpSet,
				  SmtpLcSubmission,
				  "submission"));
	break;
    case NUM_s: /* Not used */
	panic("MAILER PANIC",__FILE__,__LINE__,"Server_s",
	      "Bad idx",0);
	break;
    }

    if (!static_strings[idx])
	panic("MAILER PANIC",__FILE__,__LINE__,"Server_s",
	      "Bad idx?",0);		

    return static_strings[idx];
}




E_(free_shared_cache_f free_shared_cache)
void free_shared_cache P_((void));
void free_shared_cache() 
{
    enum server_s i;

    DPRINT(Debug,7,(&Debug, "free_shared_cache\n"));

    for (i = 0; i < NUM_s; i++)
	if (static_strings[i])
	    free_string(& static_strings[i]);
}


#if 0

/* elmrc options ---------------------------------------------------------- */

#include "../../hdrs/rc_imp.h"
#include "../../hdrs/save_opts.h"

#endif


#endif

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
